import json
import logging
from typing import Any, Optional

from antlr4 import ParserRuleContext
from antlr4.tree.Tree import ParseTree, TerminalNodeImpl

from moto.stepfunctions.parser.asl.antlr.runtime.ASLLexer import ASLLexer
from moto.stepfunctions.parser.asl.antlr.runtime.ASLParser import ASLParser
from moto.stepfunctions.parser.asl.antlr.runtime.ASLParserVisitor import (
    ASLParserVisitor,
)
from moto.stepfunctions.parser.asl.antlt4utils.antlr4utils import (
    from_string_literal,
    is_production,
    is_terminal,
)
from moto.stepfunctions.parser.asl.component.common.assign.assign_decl import AssignDecl
from moto.stepfunctions.parser.asl.component.common.assign.assign_decl_binding import (
    AssignDeclBinding,
)
from moto.stepfunctions.parser.asl.component.common.assign.assign_template_binding import (
    AssignTemplateBinding,
    AssignTemplateBindingStringExpressionSimple,
    AssignTemplateBindingValue,
)
from moto.stepfunctions.parser.asl.component.common.assign.assign_template_value import (
    AssignTemplateValue,
)
from moto.stepfunctions.parser.asl.component.common.assign.assign_template_value_array import (
    AssignTemplateValueArray,
)
from moto.stepfunctions.parser.asl.component.common.assign.assign_template_value_object import (
    AssignTemplateValueObject,
)
from moto.stepfunctions.parser.asl.component.common.assign.assign_template_value_terminal import (
    AssignTemplateValueTerminal,
    AssignTemplateValueTerminalLit,
    AssignTemplateValueTerminalStringJSONata,
)
from moto.stepfunctions.parser.asl.component.common.catch.catch_decl import CatchDecl
from moto.stepfunctions.parser.asl.component.common.catch.catcher_decl import (
    CatcherDecl,
)
from moto.stepfunctions.parser.asl.component.common.catch.catcher_props import (
    CatcherProps,
)
from moto.stepfunctions.parser.asl.component.common.comment import Comment
from moto.stepfunctions.parser.asl.component.common.error_name.custom_error_name import (
    CustomErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.error_equals_decl import (
    ErrorEqualsDecl,
)
from moto.stepfunctions.parser.asl.component.common.error_name.error_name import (
    ErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.common.flow.end import End
from moto.stepfunctions.parser.asl.component.common.flow.next import Next
from moto.stepfunctions.parser.asl.component.common.flow.start_at import StartAt
from moto.stepfunctions.parser.asl.component.common.jsonata.jsonata_template_binding import (
    JSONataTemplateBinding,
)
from moto.stepfunctions.parser.asl.component.common.jsonata.jsonata_template_value import (
    JSONataTemplateValue,
)
from moto.stepfunctions.parser.asl.component.common.jsonata.jsonata_template_value_array import (
    JSONataTemplateValueArray,
)
from moto.stepfunctions.parser.asl.component.common.jsonata.jsonata_template_value_object import (
    JSONataTemplateValueObject,
)
from moto.stepfunctions.parser.asl.component.common.jsonata.jsonata_template_value_terminal import (
    JSONataTemplateValueTerminalLit,
    JSONataTemplateValueTerminalStringJSONata,
)
from moto.stepfunctions.parser.asl.component.common.outputdecl import Output
from moto.stepfunctions.parser.asl.component.common.parargs import (
    ArgumentsJSONataTemplateValueObject,
    ArgumentsStringJSONata,
    Parameters,
    Parargs,
)
from moto.stepfunctions.parser.asl.component.common.path.input_path import InputPath
from moto.stepfunctions.parser.asl.component.common.path.items_path import ItemsPath
from moto.stepfunctions.parser.asl.component.common.path.output_path import OutputPath
from moto.stepfunctions.parser.asl.component.common.path.result_path import ResultPath
from moto.stepfunctions.parser.asl.component.common.payload.payloadvalue.payload_value import (
    PayloadValue,
)
from moto.stepfunctions.parser.asl.component.common.payload.payloadvalue.payloadarr.payload_arr import (
    PayloadArr,
)
from moto.stepfunctions.parser.asl.component.common.payload.payloadvalue.payloadbinding.payload_binding import (
    PayloadBinding,
    PayloadBindingStringExpressionSimple,
    PayloadBindingValue,
)
from moto.stepfunctions.parser.asl.component.common.payload.payloadvalue.payloadtmpl.payload_tmpl import (
    PayloadTmpl,
)
from moto.stepfunctions.parser.asl.component.common.payload.payloadvalue.payloadvaluelit.payload_value_bool import (
    PayloadValueBool,
)
from moto.stepfunctions.parser.asl.component.common.payload.payloadvalue.payloadvaluelit.payload_value_float import (
    PayloadValueFloat,
)
from moto.stepfunctions.parser.asl.component.common.payload.payloadvalue.payloadvaluelit.payload_value_int import (
    PayloadValueInt,
)
from moto.stepfunctions.parser.asl.component.common.payload.payloadvalue.payloadvaluelit.payload_value_null import (
    PayloadValueNull,
)
from moto.stepfunctions.parser.asl.component.common.payload.payloadvalue.payloadvaluelit.payload_value_str import (
    PayloadValueStr,
)
from moto.stepfunctions.parser.asl.component.common.query_language import (
    QueryLanguage,
    QueryLanguageMode,
)
from moto.stepfunctions.parser.asl.component.common.result_selector import (
    ResultSelector,
)
from moto.stepfunctions.parser.asl.component.common.retry.backoff_rate_decl import (
    BackoffRateDecl,
)
from moto.stepfunctions.parser.asl.component.common.retry.interval_seconds_decl import (
    IntervalSecondsDecl,
)
from moto.stepfunctions.parser.asl.component.common.retry.jitter_strategy_decl import (
    JitterStrategy,
    JitterStrategyDecl,
)
from moto.stepfunctions.parser.asl.component.common.retry.max_attempts_decl import (
    MaxAttemptsDecl,
)
from moto.stepfunctions.parser.asl.component.common.retry.max_delay_seconds_decl import (
    MaxDelaySecondsDecl,
)
from moto.stepfunctions.parser.asl.component.common.retry.retrier_decl import (
    RetrierDecl,
)
from moto.stepfunctions.parser.asl.component.common.retry.retrier_props import (
    RetrierProps,
)
from moto.stepfunctions.parser.asl.component.common.retry.retry_decl import RetryDecl
from moto.stepfunctions.parser.asl.component.common.string.string_expression import (
    StringContextPath,
    StringExpression,
    StringExpressionSimple,
    StringIntrinsicFunction,
    StringJSONata,
    StringJsonPath,
    StringLiteral,
    StringSampler,
    StringVariableSample,
)
from moto.stepfunctions.parser.asl.component.common.timeouts.heartbeat import (
    HeartbeatSeconds,
    HeartbeatSecondsJSONata,
    HeartbeatSecondsPath,
)
from moto.stepfunctions.parser.asl.component.common.timeouts.timeout import (
    TimeoutSeconds,
    TimeoutSecondsJSONata,
    TimeoutSecondsPath,
)
from moto.stepfunctions.parser.asl.component.component import Component
from moto.stepfunctions.parser.asl.component.program.program import Program
from moto.stepfunctions.parser.asl.component.program.states import States
from moto.stepfunctions.parser.asl.component.program.version import Version
from moto.stepfunctions.parser.asl.component.state.choice.choice_rule import (
    ChoiceRule,
)
from moto.stepfunctions.parser.asl.component.state.choice.choices_decl import (
    ChoicesDecl,
)
from moto.stepfunctions.parser.asl.component.state.choice.comparison.comparison import (
    ComparisonComposite,
    ComparisonCompositeAnd,
    ComparisonCompositeNot,
    ComparisonCompositeOr,
    ComparisonCompositeProps,
    ConditionJSONataLit,
    ConditionStringJSONata,
)
from moto.stepfunctions.parser.asl.component.state.choice.comparison.comparison_func import (
    ComparisonFunc,
    ComparisonFuncStringVariableSample,
    ComparisonFuncValue,
)
from moto.stepfunctions.parser.asl.component.state.choice.comparison.comparison_operator_type import (
    ComparisonOperatorType,
)
from moto.stepfunctions.parser.asl.component.state.choice.comparison.comparison_type import (
    Comparison,
)
from moto.stepfunctions.parser.asl.component.state.choice.comparison.comparison_variable import (
    ComparisonVariable,
)
from moto.stepfunctions.parser.asl.component.state.choice.comparison.variable import (
    Variable,
)
from moto.stepfunctions.parser.asl.component.state.choice.default_decl import (
    DefaultDecl,
)
from moto.stepfunctions.parser.asl.component.state.choice.state_choice import (
    StateChoice,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.execution_type import (
    ExecutionType,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.item_reader_decl import (
    ItemReader,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.reader_config.csv_header_location import (
    CSVHeaderLocation,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.reader_config.csv_headers import (
    CSVHeaders,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.reader_config.input_type import (
    InputType,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.reader_config.max_items_decl import (
    MaxItemsDecl,
    MaxItemsInt,
    MaxItemsPath,
    MaxItemsStringJSONata,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.reader_config.reader_config_decl import (
    ReaderConfig,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.reader_config.reader_config_props import (
    ReaderConfigProps,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_selector import (
    ItemSelector,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.items.items import (
    ItemsArray,
    ItemsJSONata,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.itemprocessor.item_processor_decl import (
    ItemProcessorDecl,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.itemprocessor.processor_config import (
    ProcessorConfig,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iterator.iterator_decl import (
    IteratorDecl,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.label import (
    Label,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.max_concurrency import (
    MaxConcurrency,
    MaxConcurrencyJSONata,
    MaxConcurrencyPath,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.mode import (
    Mode,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.result_writer.result_writer_decl import (
    ResultWriter,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.state_map import (
    StateMap,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.tolerated_failure import (
    ToleratedFailureCountInt,
    ToleratedFailureCountPath,
    ToleratedFailurePercentage,
    ToleratedFailurePercentagePath,
    ToleratedFailurePercentageStringJSONata,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_parallel.branches_decl import (
    BranchesDecl,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_parallel.state_parallel import (
    StateParallel,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.credentials import (
    Credentials,
    RoleArn,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    Resource,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.state_task_factory import (
    state_task_for,
)
from moto.stepfunctions.parser.asl.component.state.fail.cause_decl import (
    Cause,
    CausePath,
)
from moto.stepfunctions.parser.asl.component.state.fail.error_decl import (
    Error,
    ErrorPath,
)
from moto.stepfunctions.parser.asl.component.state.fail.state_fail import StateFail
from moto.stepfunctions.parser.asl.component.state.state import CommonStateField
from moto.stepfunctions.parser.asl.component.state.state_pass.result import Result
from moto.stepfunctions.parser.asl.component.state.state_pass.state_pass import (
    StatePass,
)
from moto.stepfunctions.parser.asl.component.state.state_props import StateProps
from moto.stepfunctions.parser.asl.component.state.state_succeed.state_succeed import (
    StateSucceed,
)
from moto.stepfunctions.parser.asl.component.state.state_type import StateType
from moto.stepfunctions.parser.asl.component.state.wait.state_wait import StateWait
from moto.stepfunctions.parser.asl.component.state.wait.wait_function.seconds import (
    Seconds,
    SecondsJSONata,
)
from moto.stepfunctions.parser.asl.component.state.wait.wait_function.seconds_path import (
    SecondsPath,
)
from moto.stepfunctions.parser.asl.component.state.wait.wait_function.timestamp import (
    Timestamp,
    TimestampPath,
)
from moto.stepfunctions.parser.asl.parse.intrinsic.intrinsic_parser import (
    IntrinsicParser,
)
from moto.stepfunctions.parser.asl.parse.typed_props import TypedProps

LOG = logging.getLogger(__name__)


class Preprocessor(ASLParserVisitor):
    _query_language_per_scope: list[QueryLanguage] = []

    def _get_current_query_language(self) -> QueryLanguage:
        return self._query_language_per_scope[-1]

    def _open_query_language_scope(self, parse_tree: ParseTree) -> None:
        production = is_production(parse_tree)
        if production is None:
            raise RuntimeError(
                f"Cannot expect QueryLanguage definition at depth: {parse_tree}"
            )

        # Extract the QueryLanguage declaration at this ParseTree level, if any.
        query_language = None
        for child in production.children:
            sub_production = is_production(
                child, ASLParser.RULE_top_layer_stmt
            ) or is_production(child, ASLParser.RULE_state_stmt)
            if sub_production is not None:
                child = sub_production.children[0]
            sub_production = is_production(child, ASLParser.RULE_query_language_decl)
            if sub_production is not None:
                query_language = self.visit(sub_production)
                break

        # Check this is the initial scope, if so set the initial value to the declaration or the default.
        if not self._query_language_per_scope:
            if query_language is None:
                query_language = QueryLanguage()
        # Otherwise, check for logical conflicts and add the latest or inherited value to as the next scope.
        else:
            current_query_language = self._get_current_query_language()
            if query_language is None:
                query_language = current_query_language
            if (
                current_query_language.query_language_mode == QueryLanguageMode.JSONata
                and query_language.query_language_mode == QueryLanguageMode.JSONPath
            ):
                raise ValueError(
                    f"Cannot downgrade from JSONata context to a JSONPath context at: {parse_tree}"
                )

        self._query_language_per_scope.append(query_language)

    def _close_query_language_scope(self) -> None:
        self._query_language_per_scope.pop()

    def _is_query_language(self, query_language_mode: QueryLanguageMode) -> bool:
        current_query_language = self._get_current_query_language()
        return current_query_language.query_language_mode == query_language_mode

    def _raise_if_query_language_is_not(
        self, query_language_mode: QueryLanguageMode, ctx: ParserRuleContext
    ) -> None:
        if not self._is_query_language(query_language_mode=query_language_mode):
            raise ValueError(
                f"Unsupported declaration in QueryLanguage={query_language_mode} block: {ctx.getText()}"
            )

    @staticmethod
    def _inner_string_of(parser_rule_context: ParserRuleContext) -> Optional[str]:
        if is_terminal(parser_rule_context, ASLLexer.NULL):
            return None
        inner_str = parser_rule_context.getText()
        if inner_str.startswith('"') and inner_str.endswith('"'):
            inner_str = inner_str[1:-1]
        return inner_str

    def _inner_jsonata_expr(self, ctx: ParserRuleContext) -> str:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONata, ctx=ctx
        )
        inner_string_value = from_string_literal(parser_rule_context=ctx)
        # Strip the start and end jsonata symbols {%<body>%}
        expression_body = inner_string_value[2:-2]
        # Often leading and trailing spaces are used around the body: remove.
        expression = expression_body.strip()
        return expression

    def visitComment_decl(self, ctx: ASLParser.Comment_declContext) -> Comment:
        inner_str = self._inner_string_of(parser_rule_context=ctx.string_literal())
        return Comment(comment=inner_str)

    def visitVersion_decl(self, ctx: ASLParser.Version_declContext) -> Version:
        version_str = self._inner_string_of(parser_rule_context=ctx.string_literal())
        return Version(version=version_str)

    def visitStartat_decl(self, ctx: ASLParser.Startat_declContext) -> StartAt:
        inner_str = self._inner_string_of(parser_rule_context=ctx.string_literal())
        return StartAt(start_at_name=inner_str)

    def visitStates_decl(self, ctx: ASLParser.States_declContext) -> States:
        states = States()
        for child in ctx.children:
            cmp: Optional[Component] = self.visit(child)
            if isinstance(cmp, CommonStateField):
                # TODO move check to setter or checker layer?
                if cmp.name in states.states:
                    raise ValueError(f"State redefinition {child.getText()}")
                states.states[cmp.name] = cmp
        return states

    def visitType_decl(self, ctx: ASLParser.Type_declContext) -> StateType:
        return self.visit(ctx.state_type())

    def visitState_type(self, ctx: ASLParser.State_typeContext) -> StateType:
        state_type: int = ctx.children[0].symbol.type
        return StateType(state_type)

    def visitResource_decl(self, ctx: ASLParser.Resource_declContext) -> Resource:
        inner_str = self._inner_string_of(parser_rule_context=ctx.string_literal())
        return Resource.from_resource_arn(inner_str)

    def visitEnd_decl(self, ctx: ASLParser.End_declContext) -> End:
        bool_child: ParseTree = ctx.children[-1]
        bool_term: Optional[TerminalNodeImpl] = is_terminal(bool_child)
        if bool_term is None:
            raise ValueError(
                f"Could not derive End from declaration context '{ctx.getText()}'"
            )
        bool_term_rule: int = bool_term.getSymbol().type
        is_end = bool_term_rule == ASLLexer.TRUE
        return End(is_end=is_end)

    def visitNext_decl(self, ctx: ASLParser.Next_declContext) -> Next:
        inner_str = self._inner_string_of(parser_rule_context=ctx.string_literal())
        return Next(name=inner_str)

    def visitResult_path_decl(
        self, ctx: ASLParser.Result_path_declContext
    ) -> ResultPath:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        inner_str = self._inner_string_of(parser_rule_context=ctx.children[-1])
        return ResultPath(result_path_src=inner_str)

    def visitInput_path_decl(self, ctx: ASLParser.Input_path_declContext) -> InputPath:
        string_sampler: Optional[StringSampler] = None
        if not is_terminal(pt=ctx.children[-1], token_type=ASLLexer.NULL):
            string_sampler: StringSampler = self.visitString_sampler(
                ctx.string_sampler()
            )
        return InputPath(string_sampler=string_sampler)

    def visitOutput_path_decl(
        self, ctx: ASLParser.Output_path_declContext
    ) -> OutputPath:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        string_sampler: Optional[StringSampler] = None
        if is_production(ctx.children[-1], ASLParser.RULE_string_sampler):
            string_sampler: StringSampler = self.visitString_sampler(ctx.children[-1])
        return OutputPath(string_sampler=string_sampler)

    def visitResult_decl(self, ctx: ASLParser.Result_declContext) -> Result:
        json_decl = ctx.json_value_decl()
        json_str: str = json_decl.getText()
        json_obj: json = json.loads(json_str)
        return Result(result_obj=json_obj)

    def visitParameters_decl(self, ctx: ASLParser.Parameters_declContext) -> Parameters:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        payload_tmpl: PayloadTmpl = self.visit(ctx.payload_tmpl_decl())
        return Parameters(payload_tmpl=payload_tmpl)

    def visitTimeout_seconds_int(
        self, ctx: ASLParser.Timeout_seconds_intContext
    ) -> TimeoutSeconds:
        seconds = int(ctx.INT().getText())
        return TimeoutSeconds(timeout_seconds=seconds)

    def visitTimeout_seconds_jsonata(
        self, ctx: ASLParser.Timeout_seconds_jsonataContext
    ) -> TimeoutSecondsJSONata:
        string_jsonata: StringJSONata = self.visitString_jsonata(ctx.string_jsonata())
        return TimeoutSecondsJSONata(string_jsonata=string_jsonata)

    def visitTimeout_seconds_path(
        self, ctx: ASLParser.Timeout_seconds_pathContext
    ) -> TimeoutSecondsPath:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        string_sampler: StringSampler = self.visitString_sampler(ctx.string_sampler())
        return TimeoutSecondsPath(string_sampler=string_sampler)

    def visitHeartbeat_seconds_int(
        self, ctx: ASLParser.Heartbeat_seconds_intContext
    ) -> HeartbeatSeconds:
        seconds = int(ctx.INT().getText())
        return HeartbeatSeconds(heartbeat_seconds=seconds)

    def visitHeartbeat_seconds_jsonata(
        self, ctx: ASLParser.Heartbeat_seconds_jsonataContext
    ) -> HeartbeatSecondsJSONata:
        string_jsonata: StringJSONata = self.visitString_jsonata(ctx.string_jsonata())
        return HeartbeatSecondsJSONata(string_jsonata=string_jsonata)

    def visitHeartbeat_seconds_path(
        self, ctx: ASLParser.Heartbeat_seconds_pathContext
    ) -> HeartbeatSecondsPath:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        string_sampler: StringSampler = self.visitString_sampler(ctx.string_sampler())
        return HeartbeatSecondsPath(string_sampler=string_sampler)

    def visitResult_selector_decl(
        self, ctx: ASLParser.Result_selector_declContext
    ) -> ResultSelector:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        payload_tmpl: PayloadTmpl = self.visit(ctx.payload_tmpl_decl())
        return ResultSelector(payload_tmpl=payload_tmpl)

    def visitBranches_decl(self, ctx: ASLParser.Branches_declContext) -> BranchesDecl:
        programs: list[Program] = []
        for child in ctx.children:
            cmp: Optional[Component] = self.visit(child)
            if isinstance(cmp, Program):
                programs.append(cmp)
        return BranchesDecl(programs=programs)

    def visitState_decl_body(self, ctx: ASLParser.State_decl_bodyContext) -> StateProps:
        self._open_query_language_scope(ctx)
        state_props = StateProps()
        for child in ctx.children:
            cmp: Optional[Component] = self.visit(child)
            state_props.add(cmp)
        if state_props.get(QueryLanguage) is None:
            state_props.add(self._get_current_query_language())
        self._close_query_language_scope()
        return state_props

    def visitState_decl(self, ctx: ASLParser.State_declContext) -> CommonStateField:
        state_name = self._inner_string_of(parser_rule_context=ctx.string_literal())
        state_props: StateProps = self.visit(ctx.state_decl_body())
        state_props.name = state_name
        common_state_field = self._common_state_field_of(state_props=state_props)
        return common_state_field

    @staticmethod
    def _common_state_field_of(state_props: StateProps) -> CommonStateField:
        # TODO: use subtype loading strategy.
        state_type = state_props.get(StateType)
        if state_type == StateType.Task:
            resource: Resource = state_props.get(Resource)
            state = state_task_for(resource)
        elif state_type == StateType.Pass:
            state = StatePass()
        elif state_type == StateType.Choice:
            state = StateChoice()
        elif state_type == StateType.Fail:
            state = StateFail()
        elif state_type == StateType.Succeed:
            state = StateSucceed()
        elif state_type == StateType.Wait:
            state = StateWait()
        elif state_type == StateType.Map:
            state = StateMap()
        elif state_type == StateType.Parallel:
            state = StateParallel()
        elif state_type is None:
            raise TypeError("No Type declaration for State in context.")
        else:
            raise TypeError(
                f"Unknown StateType value '{unknown}' in StateProps object in context."  # noqa
            )
        state.from_state_props(state_props)
        return state

    def visitCondition_lit(
        self, ctx: ASLParser.Condition_litContext
    ) -> ConditionJSONataLit:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONata, ctx=ctx
        )
        bool_child: ParseTree = ctx.children[-1]
        bool_term: Optional[TerminalNodeImpl] = is_terminal(bool_child)
        if bool_term is None:
            raise ValueError(
                f"Could not derive boolean literal from declaration context '{ctx.getText()}'."
            )
        bool_term_rule: int = bool_term.getSymbol().type
        bool_val: bool = bool_term_rule == ASLLexer.TRUE
        return ConditionJSONataLit(literal=bool_val)

    def visitCondition_string_jsonata(
        self, ctx: ASLParser.Condition_string_jsonataContext
    ) -> ConditionStringJSONata:
        string_jsonata: StringJSONata = self.visitString_jsonata(
            ctx=ctx.string_jsonata()
        )
        return ConditionStringJSONata(string_jsonata=string_jsonata)

    def visitVariable_decl(self, ctx: ASLParser.Variable_declContext) -> Variable:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        string_sampler: StringSampler = self.visitString_sampler(
            ctx=ctx.string_sampler()
        )
        return Variable(string_sampler=string_sampler)

    def visitComparison_op(
        self, ctx: ASLParser.Comparison_opContext
    ) -> ComparisonOperatorType:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        try:
            operator_type: int = ctx.children[0].symbol.type
            return ComparisonOperatorType(operator_type)
        except Exception:
            raise ValueError(
                f"Could not derive ComparisonOperator from context '{ctx.getText()}'."
            )

    def visitComparison_func_value(
        self, ctx: ASLParser.Comparison_func_valueContext
    ) -> ComparisonFuncValue:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        comparison_op: ComparisonOperatorType = self.visit(ctx.comparison_op())
        json_decl = ctx.json_value_decl()
        json_str: str = json_decl.getText()
        json_obj: Any = json.loads(json_str)
        return ComparisonFuncValue(operator_type=comparison_op, value=json_obj)

    def visitComparison_func_string_variable_sample(
        self, ctx: ASLParser.Comparison_func_string_variable_sampleContext
    ) -> ComparisonFuncStringVariableSample:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        comparison_op: ComparisonOperatorType = self.visit(ctx.comparison_op())
        string_variable_sample: StringVariableSample = self.visitString_variable_sample(
            ctx.string_variable_sample()
        )
        return ComparisonFuncStringVariableSample(
            operator_type=comparison_op, string_variable_sample=string_variable_sample
        )

    def visitDefault_decl(self, ctx: ASLParser.Default_declContext) -> DefaultDecl:
        state_name = self._inner_string_of(parser_rule_context=ctx.string_literal())
        return DefaultDecl(state_name=state_name)

    def visitChoice_operator(
        self, ctx: ASLParser.Choice_operatorContext
    ) -> ComparisonComposite.ChoiceOp:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        pt: Optional[TerminalNodeImpl] = is_terminal(ctx.children[0])
        if not pt:
            raise ValueError(
                f"Could not derive ChoiceOperator in block '{ctx.getText()}'."
            )
        return ComparisonComposite.ChoiceOp(pt.symbol.type)

    def visitComparison_composite(
        self, ctx: ASLParser.Comparison_compositeContext
    ) -> ComparisonComposite:
        choice_op: ComparisonComposite.ChoiceOp = self.visit(ctx.choice_operator())
        rules: list[ChoiceRule] = []
        for child in ctx.children[1:]:
            cmp: Optional[Component] = self.visit(child)
            if not cmp:
                continue
            elif isinstance(cmp, ChoiceRule):
                rules.append(cmp)

        if choice_op == ComparisonComposite.ChoiceOp.Not:
            if len(rules) != 1:
                raise ValueError(
                    f"ComparisonCompositeNot must carry only one ComparisonCompositeStmt in: '{ctx.getText()}'."
                )
            return ComparisonCompositeNot(rule=rules[0])
        elif choice_op == ComparisonComposite.ChoiceOp.And:
            return ComparisonCompositeAnd(rules=rules)
        elif choice_op == ComparisonComposite.ChoiceOp.Or:
            return ComparisonCompositeOr(rules=rules)

    def visitChoice_rule_comparison_composite(
        self, ctx: ASLParser.Choice_rule_comparison_compositeContext
    ) -> ChoiceRule:
        composite_stmts = ComparisonCompositeProps()
        for child in ctx.children:
            cmp: Optional[Component] = self.visit(child)
            composite_stmts.add(cmp)
        return ChoiceRule(
            comparison=composite_stmts.get(
                typ=ComparisonComposite,
                raise_on_missing=ValueError(
                    f"Expecting a 'ComparisonComposite' definition at '{ctx.getText()}'."
                ),
            ),
            next_stmt=composite_stmts.get(Next),
            comment=composite_stmts.get(Comment),
            assign=composite_stmts.get(AssignDecl),
            output=composite_stmts.get(Output),
        )

    def visitChoice_rule_comparison_variable(
        self, ctx: ASLParser.Choice_rule_comparison_variableContext
    ) -> ChoiceRule:
        comparison_stmts = StateProps()
        for child in ctx.children:
            cmp: Optional[Component] = self.visit(child)
            comparison_stmts.add(cmp)
        if self._is_query_language(query_language_mode=QueryLanguageMode.JSONPath):
            variable: Variable = comparison_stmts.get(
                typ=Variable,
                raise_on_missing=ValueError(
                    f"Expected a Variable declaration in '{ctx.getText()}'."
                ),
            )
            comparison_func: Comparison = comparison_stmts.get(
                typ=Comparison,
                raise_on_missing=ValueError(
                    f"Expected a ComparisonFunction declaration in '{ctx.getText()}'."
                ),
            )
            if not isinstance(comparison_func, ComparisonFunc):
                raise ValueError(
                    f"Expected a ComparisonFunction declaration in '{ctx.getText()}'"
                )
            comparison_variable = ComparisonVariable(
                variable=variable, func=comparison_func
            )
            return ChoiceRule(
                comparison=comparison_variable,
                next_stmt=comparison_stmts.get(Next),
                comment=comparison_stmts.get(Comment),
                assign=None,
                output=None,
            )
        else:
            condition: Comparison = comparison_stmts.get(
                typ=Comparison,
                raise_on_missing=ValueError(
                    f"Expected a Condition declaration in '{ctx.getText()}'"
                ),
            )
            return ChoiceRule(
                comparison=condition,
                next_stmt=comparison_stmts.get(Next),
                comment=comparison_stmts.get(Comment),
                assign=comparison_stmts.get(AssignDecl),
                output=comparison_stmts.get(Output),
            )

    def visitChoices_decl(self, ctx: ASLParser.Choices_declContext) -> ChoicesDecl:
        rules: list[ChoiceRule] = []
        for child in ctx.children:
            cmp: Optional[Component] = self.visit(child)
            if not cmp:
                continue
            elif isinstance(cmp, ChoiceRule):
                rules.append(cmp)
        return ChoicesDecl(rules=rules)

    def visitError(self, ctx: ASLParser.ErrorContext) -> Error:
        string_expression: StringExpression = self.visit(ctx.children[-1])
        return Error(string_expression=string_expression)

    def visitError_path(self, ctx: ASLParser.Error_pathContext) -> ErrorPath:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        string_expression: StringExpression = self.visit(ctx.children[-1])
        return ErrorPath(string_expression=string_expression)

    def visitCause(self, ctx: ASLParser.CauseContext) -> Cause:
        string_expression: StringExpression = self.visit(ctx.children[-1])
        return Cause(string_expression=string_expression)

    def visitCause_path(self, ctx: ASLParser.Cause_pathContext) -> CausePath:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        string_expression: StringExpression = self.visit(ctx.children[-1])
        return CausePath(string_expression=string_expression)

    def visitRole_arn(self, ctx: ASLParser.Role_arnContext) -> RoleArn:
        string_expression: StringExpression = self.visit(ctx.children[-1])
        return RoleArn(string_expression=string_expression)

    def visitRole_path(self, ctx: ASLParser.Role_pathContext) -> RoleArn:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        string_expression_simple: StringExpressionSimple = (
            self.visitString_expression_simple(ctx=ctx.string_expression_simple())
        )
        return RoleArn(string_expression=string_expression_simple)

    def visitCredentials_decl(
        self, ctx: ASLParser.Credentials_declContext
    ) -> Credentials:
        role_arn: RoleArn = self.visit(ctx.role_arn_decl())
        return Credentials(role_arn=role_arn)

    def visitSeconds_int(self, ctx: ASLParser.Seconds_intContext) -> Seconds:
        return Seconds(seconds=int(ctx.INT().getText()))

    def visitSeconds_jsonata(
        self, ctx: ASLParser.Seconds_jsonataContext
    ) -> SecondsJSONata:
        string_jsonata: StringJSONata = self.visitString_jsonata(ctx.string_jsonata())
        return SecondsJSONata(string_jsonata=string_jsonata)

    def visitSeconds_path(self, ctx: ASLParser.Seconds_pathContext) -> SecondsPath:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        string_sampler: StringSampler = self.visitString_sampler(
            ctx=ctx.string_sampler()
        )
        return SecondsPath(string_sampler=string_sampler)

    def visitItems_path_decl(self, ctx: ASLParser.Items_path_declContext) -> ItemsPath:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        string_sampler: StringSampler = self.visitString_sampler(ctx.string_sampler())
        return ItemsPath(string_sampler=string_sampler)

    def visitMax_concurrency_int(
        self, ctx: ASLParser.Max_concurrency_intContext
    ) -> MaxConcurrency:
        return MaxConcurrency(num=int(ctx.INT().getText()))

    def visitMax_concurrency_jsonata(
        self, ctx: ASLParser.Max_concurrency_jsonataContext
    ) -> MaxConcurrencyJSONata:
        string_jsonata: StringJSONata = self.visitString_jsonata(ctx.string_jsonata())
        return MaxConcurrencyJSONata(string_jsonata=string_jsonata)

    def visitMax_concurrency_path(
        self, ctx: ASLParser.Max_concurrency_pathContext
    ) -> MaxConcurrencyPath:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        string_sampler: StringSampler = self.visitString_sampler(ctx.string_sampler())
        return MaxConcurrencyPath(string_sampler=string_sampler)

    def visitMode_decl(self, ctx: ASLParser.Mode_declContext) -> Mode:
        mode_type: int = self.visit(ctx.mode_type())
        return Mode(mode_type)

    def visitMode_type(self, ctx: ASLParser.Mode_typeContext) -> int:
        return ctx.children[0].symbol.type

    def visitExecution_decl(
        self, ctx: ASLParser.Execution_declContext
    ) -> ExecutionType:
        execution_type: int = self.visit(ctx.execution_type())
        return ExecutionType(execution_type)

    def visitExecution_type(self, ctx: ASLParser.Execution_typeContext) -> int:
        return ctx.children[0].symbol.type

    def visitTimestamp(self, ctx: ASLParser.TimestampContext) -> Timestamp:
        string: StringExpression = self.visit(ctx.children[-1])
        return Timestamp(string=string)

    def visitTimestamp_path(
        self, ctx: ASLParser.Timestamp_pathContext
    ) -> TimestampPath:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        string_sampler: StringSampler = self.visitString_sampler(ctx.string_sampler())
        return TimestampPath(string=string_sampler)

    def visitProcessor_config_decl(
        self, ctx: ASLParser.Processor_config_declContext
    ) -> ProcessorConfig:
        props = TypedProps()
        for child in ctx.children:
            cmp = self.visit(child)
            props.add(cmp)
        return ProcessorConfig(
            mode=props.get(typ=Mode) or ProcessorConfig.DEFAULT_MODE,
            execution_type=props.get(typ=ExecutionType)
            or ProcessorConfig.DEFAULT_EXECUTION_TYPE,
        )

    def visitItem_processor_item(
        self, ctx: ASLParser.Item_processor_itemContext
    ) -> Component:
        return self.visit(ctx.children[0])

    def visitItem_processor_decl(
        self, ctx: ASLParser.Item_processor_declContext
    ) -> ItemProcessorDecl:
        props = TypedProps()
        for child in ctx.children:
            cmp = self.visit(child)
            props.add(cmp)
        return ItemProcessorDecl(
            query_language=props.get(QueryLanguage) or QueryLanguage(),
            start_at=props.get(
                typ=StartAt,
                raise_on_missing=ValueError(
                    f"Expected a StartAt declaration at '{ctx.getText()}'."
                ),
            ),
            states=props.get(
                typ=States,
                raise_on_missing=ValueError(
                    f"Expected a States declaration at '{ctx.getText()}'."
                ),
            ),
            comment=props.get(typ=Comment),
            processor_config=props.get(typ=ProcessorConfig) or ProcessorConfig(),
        )

    def visitIterator_decl(self, ctx: ASLParser.Iterator_declContext) -> IteratorDecl:
        props = TypedProps()
        for child in ctx.children:
            cmp = self.visit(child)
            props.add(cmp)
        return IteratorDecl(
            comment=props.get(typ=Comment),
            query_language=self._get_current_query_language(),
            start_at=props.get(
                typ=StartAt,
                raise_on_missing=ValueError(
                    f"Expected a StartAt declaration at '{ctx.getText()}'."
                ),
            ),
            states=props.get(
                typ=States,
                raise_on_missing=ValueError(
                    f"Expected a States declaration at '{ctx.getText()}'."
                ),
            ),
            processor_config=props.get(typ=ProcessorConfig) or ProcessorConfig(),
        )

    def visitItem_selector_decl(
        self, ctx: ASLParser.Item_selector_declContext
    ) -> ItemSelector:
        template_value_object = self.visitAssign_template_value_object(
            ctx=ctx.assign_template_value_object()
        )
        return ItemSelector(template_value_object=template_value_object)

    def visitItem_reader_decl(
        self, ctx: ASLParser.Item_reader_declContext
    ) -> ItemReader:
        props = StateProps()
        for child in ctx.children[3:-1]:
            cmp = self.visit(child)
            props.add(cmp)
        resource: Resource = props.get(
            typ=Resource,
            raise_on_missing=ValueError(
                f"Expected a Resource declaration at '{ctx.getText()}'."
            ),
        )
        return ItemReader(
            resource=resource,
            parargs=props.get(Parargs),
            reader_config=props.get(ReaderConfig),
        )

    def visitReader_config_decl(
        self, ctx: ASLParser.Reader_config_declContext
    ) -> ReaderConfig:
        props = ReaderConfigProps()
        for child in ctx.children:
            cmp = self.visit(child)
            props.add(cmp)
        return ReaderConfig(
            input_type=props.get(
                typ=InputType,
                raise_on_missing=ValueError(
                    f"Expected a InputType declaration at '{ctx.getText()}'."
                ),
            ),
            max_items_decl=props.get(typ=MaxItemsDecl),
            csv_header_location=props.get(CSVHeaderLocation),
            csv_headers=props.get(CSVHeaders),
        )

    def visitInput_type_decl(self, ctx: ASLParser.Input_type_declContext) -> InputType:
        input_type = self._inner_string_of(ctx.string_literal())
        return InputType(input_type=input_type)

    def visitCsv_header_location_decl(
        self, ctx: ASLParser.Csv_header_location_declContext
    ) -> CSVHeaderLocation:
        value = self._inner_string_of(ctx.string_literal())
        return CSVHeaderLocation(csv_header_location_value=value)

    def visitCsv_headers_decl(
        self, ctx: ASLParser.Csv_headers_declContext
    ) -> CSVHeaders:
        csv_headers: list[str] = []
        for child in ctx.children[3:-1]:
            maybe_str = is_production(
                pt=child, rule_index=ASLParser.RULE_string_literal
            )
            if maybe_str is not None:
                csv_headers.append(self._inner_string_of(maybe_str))
        # TODO: check for empty headers behaviour.
        return CSVHeaders(header_names=csv_headers)

    def visitMax_items_path(self, ctx: ASLParser.Max_items_pathContext) -> MaxItemsPath:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        string_sampler: StringSampler = self.visitString_sampler(
            ctx=ctx.string_sampler()
        )
        return MaxItemsPath(string_sampler=string_sampler)

    def visitMax_items_int(self, ctx: ASLParser.Max_items_intContext) -> MaxItemsInt:
        return MaxItemsInt(max_items=int(ctx.INT().getText()))

    def visitMax_items_string_jsonata(
        self, ctx: ASLParser.Max_items_string_jsonataContext
    ) -> MaxItemsStringJSONata:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONata, ctx=ctx
        )
        string_jsonata: StringJSONata = self.visitString_jsonata(ctx.string_jsonata())
        return MaxItemsStringJSONata(string_jsonata=string_jsonata)

    def visitTolerated_failure_count_int(
        self, ctx: ASLParser.Tolerated_failure_count_intContext
    ) -> ToleratedFailureCountInt:
        LOG.warning(
            "ToleratedFailureCount declarations currently have no effect on the program evaluation."
        )
        count = int(ctx.INT().getText())
        return ToleratedFailureCountInt(tolerated_failure_count=count)

    def visitTolerated_failure_count_string_jsonata(
        self, ctx: ASLParser.Tolerated_failure_count_string_jsonataContext
    ) -> ToleratedFailurePercentageStringJSONata:
        LOG.warning(
            "ToleratedFailureCount declarations currently have no effect on the program evaluation."
        )
        string_jsonata: StringJSONata = self.visitString_jsonata(
            ctx=ctx.string_jsonata()
        )
        return ToleratedFailurePercentageStringJSONata(string_jsonata=string_jsonata)

    def visitTolerated_failure_count_path(
        self, ctx: ASLParser.Tolerated_failure_count_pathContext
    ) -> ToleratedFailureCountPath:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        LOG.warning(
            "ToleratedFailureCountPath declarations currently have no effect on the program evaluation."
        )
        string_sampler: StringSampler = self.visitString_sampler(ctx.string_sampler())
        return ToleratedFailureCountPath(string_sampler=string_sampler)

    def visitTolerated_failure_percentage_number(
        self, ctx: ASLParser.Tolerated_failure_percentage_numberContext
    ) -> ToleratedFailurePercentage:
        LOG.warning(
            "ToleratedFailurePercentage declarations currently have no effect on the program evaluation."
        )
        percentage = float(ctx.NUMBER().getText())
        return ToleratedFailurePercentage(tolerated_failure_percentage=percentage)

    def visitTolerated_failure_percentage_string_jsonata(
        self, ctx: ASLParser.Tolerated_failure_percentage_string_jsonataContext
    ) -> ToleratedFailurePercentageStringJSONata:
        LOG.warning(
            "ToleratedFailurePercentage declarations currently have no effect on the program evaluation."
        )
        string_jsonata: StringJSONata = self.visitString_jsonata(
            ctx=ctx.string_jsonata()
        )
        return ToleratedFailurePercentageStringJSONata(string_jsonata=string_jsonata)

    def visitTolerated_failure_percentage_path(
        self, ctx: ASLParser.Tolerated_failure_percentage_pathContext
    ) -> ToleratedFailurePercentagePath:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONPath, ctx=ctx
        )
        LOG.warning(
            "ToleratedFailurePercentagePath declarations currently have no effect on the program evaluation."
        )
        string_sampler: StringSampler = self.visitString_sampler(ctx.string_sampler())
        return ToleratedFailurePercentagePath(string_sampler=string_sampler)

    def visitLabel_decl(self, ctx: ASLParser.Label_declContext) -> Label:
        label = self._inner_string_of(parser_rule_context=ctx.string_literal())
        return Label(label=label)

    def visitResult_writer_decl(
        self, ctx: ASLParser.Result_writer_declContext
    ) -> ResultWriter:
        props = StateProps()
        for child in ctx.children[3:-1]:
            cmp = self.visit(child)
            props.add(cmp)
        resource: Resource = props.get(
            typ=Resource,
            raise_on_missing=ValueError(
                f"Expected a Resource declaration at '{ctx.getText()}'."
            ),
        )
        # TODO: add tests for arguments in jsonata blocks using result writer
        parargs: Parargs = props.get(
            typ=Parargs,
            raise_on_missing=ValueError(
                f"Expected a Parameters/Arguments declaration at '{ctx.getText()}'."
            ),
        )
        return ResultWriter(resource=resource, parargs=parargs)

    def visitRetry_decl(self, ctx: ASLParser.Retry_declContext) -> RetryDecl:
        retriers: list[RetrierDecl] = []
        for child in ctx.children:
            cmp: Optional[Component] = self.visit(child)
            if isinstance(cmp, RetrierDecl):
                retriers.append(cmp)
        return RetryDecl(retriers=retriers)

    def visitRetrier_decl(self, ctx: ASLParser.Retrier_declContext) -> RetrierDecl:
        props = RetrierProps()
        for child in ctx.children:
            cmp: Optional[Component] = self.visit(child)
            props.add(cmp)
        return RetrierDecl.from_retrier_props(props=props)

    def visitRetrier_stmt(self, ctx: ASLParser.Retrier_stmtContext):
        return self.visit(ctx.children[0])

    def visitError_equals_decl(
        self, ctx: ASLParser.Error_equals_declContext
    ) -> ErrorEqualsDecl:
        error_names: list[ErrorName] = []
        for child in ctx.children:
            cmp = self.visit(child)
            if isinstance(cmp, ErrorName):
                error_names.append(cmp)
        return ErrorEqualsDecl(error_names=error_names)

    def visitError_name(self, ctx: ASLParser.Error_nameContext) -> ErrorName:
        pt = ctx.children[0]

        # Case: StatesErrorName.
        prc: Optional[ParserRuleContext] = is_production(
            pt=pt, rule_index=ASLParser.RULE_states_error_name
        )
        if prc:
            return self.visit(prc)

        # Case CustomErrorName.
        error_name = self._inner_string_of(parser_rule_context=ctx.string_literal())
        return CustomErrorName(error_name=error_name)

    def visitStates_error_name(
        self, ctx: ASLParser.States_error_nameContext
    ) -> StatesErrorName:
        pt: Optional[TerminalNodeImpl] = is_terminal(ctx.children[0])
        if not pt:
            raise ValueError(f"Could not derive ErrorName in block '{ctx.getText()}'.")
        states_error_name_type = StatesErrorNameType(pt.symbol.type)
        return StatesErrorName(states_error_name_type)

    def visitInterval_seconds_decl(
        self, ctx: ASLParser.Interval_seconds_declContext
    ) -> IntervalSecondsDecl:
        return IntervalSecondsDecl(seconds=int(ctx.INT().getText()))

    def visitMax_attempts_decl(
        self, ctx: ASLParser.Max_attempts_declContext
    ) -> MaxAttemptsDecl:
        return MaxAttemptsDecl(attempts=int(ctx.INT().getText()))

    def visitBackoff_rate_decl(
        self, ctx: ASLParser.Backoff_rate_declContext
    ) -> BackoffRateDecl:
        return BackoffRateDecl(rate=float(ctx.children[-1].getText()))

    def visitMax_delay_seconds_decl(
        self, ctx: ASLParser.Max_delay_seconds_declContext
    ) -> MaxDelaySecondsDecl:
        return MaxDelaySecondsDecl(max_delays_seconds=int(ctx.INT().getText()))

    def visitJitter_strategy_decl(
        self, ctx: ASLParser.Jitter_strategy_declContext
    ) -> JitterStrategyDecl:
        last_child: ParseTree = ctx.children[-1]
        strategy_child: Optional[TerminalNodeImpl] = is_terminal(last_child)
        strategy_value = strategy_child.getSymbol().type
        jitter_strategy = JitterStrategy(strategy_value)
        return JitterStrategyDecl(jitter_strategy=jitter_strategy)

    def visitCatch_decl(self, ctx: ASLParser.Catch_declContext) -> CatchDecl:
        catchers: list[CatcherDecl] = []
        for child in ctx.children:
            cmp: Optional[Component] = self.visit(child)
            if isinstance(cmp, CatcherDecl):
                catchers.append(cmp)
        return CatchDecl(catchers=catchers)

    def visitCatcher_decl(self, ctx: ASLParser.Catcher_declContext) -> CatcherDecl:
        props = CatcherProps()
        for child in ctx.children:
            cmp: Optional[Component] = self.visit(child)
            props.add(cmp)
        if self._is_query_language(QueryLanguageMode.JSONPath) and not props.get(
            ResultPath
        ):
            props.add(CatcherDecl.DEFAULT_RESULT_PATH)
        return CatcherDecl.from_catcher_props(props=props)

    def visitPayload_value_float(
        self, ctx: ASLParser.Payload_value_floatContext
    ) -> PayloadValueFloat:
        return PayloadValueFloat(val=float(ctx.NUMBER().getText()))

    def visitPayload_value_int(
        self, ctx: ASLParser.Payload_value_intContext
    ) -> PayloadValueInt:
        return PayloadValueInt(val=int(ctx.INT().getText()))

    def visitPayload_value_bool(
        self, ctx: ASLParser.Payload_value_boolContext
    ) -> PayloadValueBool:
        bool_child: ParseTree = ctx.children[0]
        bool_term: Optional[TerminalNodeImpl] = is_terminal(bool_child)
        if bool_term is None:
            raise ValueError(
                f"Could not derive PayloadValueBool from declaration context '{ctx.getText()}'."
            )
        bool_term_rule: int = bool_term.getSymbol().type
        bool_val: bool = bool_term_rule == ASLLexer.TRUE
        return PayloadValueBool(val=bool_val)

    def visitPayload_value_null(
        self, ctx: ASLParser.Payload_value_nullContext
    ) -> PayloadValueNull:
        return PayloadValueNull()

    def visitPayload_value_str(
        self, ctx: ASLParser.Payload_value_strContext
    ) -> PayloadValueStr:
        string_literal: StringLiteral = self.visitString_literal(
            ctx=ctx.string_literal()
        )
        return PayloadValueStr(val=string_literal.literal_value)

    def visitPayload_binding_sample(
        self, ctx: ASLParser.Payload_binding_sampleContext
    ) -> PayloadBindingStringExpressionSimple:
        string_dollar: str = self._inner_string_of(
            parser_rule_context=ctx.STRINGDOLLAR()
        )
        field = string_dollar[:-2]
        string_expression_simple: StringExpressionSimple = (
            self.visitString_expression_simple(ctx.string_expression_simple())
        )
        return PayloadBindingStringExpressionSimple(
            field=field, string_expression_simple=string_expression_simple
        )

    def visitPayload_binding_value(
        self, ctx: ASLParser.Payload_binding_valueContext
    ) -> PayloadBindingValue:
        string_literal: StringLiteral = self.visitString_literal(
            ctx=ctx.string_literal()
        )
        payload_value: PayloadValue = self.visit(ctx.payload_value_decl())
        return PayloadBindingValue(
            field=string_literal.literal_value, payload_value=payload_value
        )

    def visitPayload_arr_decl(
        self, ctx: ASLParser.Payload_arr_declContext
    ) -> PayloadArr:
        payload_values: list[PayloadValue] = []
        for child in ctx.children:
            cmp: Optional[Component] = self.visit(child)
            if isinstance(cmp, PayloadValue):
                payload_values.append(cmp)
        return PayloadArr(payload_values=payload_values)

    def visitPayload_tmpl_decl(
        self, ctx: ASLParser.Payload_tmpl_declContext
    ) -> PayloadTmpl:
        payload_bindings: list[PayloadBinding] = []
        for child in ctx.children:
            cmp: Optional[Component] = self.visit(child)
            if isinstance(cmp, PayloadBinding):
                payload_bindings.append(cmp)
        return PayloadTmpl(payload_bindings=payload_bindings)

    def visitPayload_value_decl(
        self, ctx: ASLParser.Payload_value_declContext
    ) -> PayloadValue:
        value = ctx.children[0]
        return self.visit(value)

    def visitProgram_decl(self, ctx: ASLParser.Program_declContext) -> Program:
        self._open_query_language_scope(ctx)
        props = TypedProps()
        for child in ctx.children:
            cmp: Optional[Component] = self.visit(child)
            props.add(cmp)
        if props.get(QueryLanguage) is None:
            props.add(self._get_current_query_language())
        program = Program(
            query_language=props.get(typ=QueryLanguage) or QueryLanguage(),
            start_at=props.get(
                typ=StartAt,
                raise_on_missing=ValueError(
                    f"No '{StartAt}' definition for Program in context: '{ctx.getText()}'."
                ),
            ),
            states=props.get(
                typ=States,
                raise_on_missing=ValueError(
                    f"No '{States}' definition for Program in context: '{ctx.getText()}'."
                ),
            ),
            timeout_seconds=props.get(TimeoutSeconds),
            comment=props.get(typ=Comment),
            version=props.get(typ=Version),
        )
        self._close_query_language_scope()
        return program

    def visitState_machine(self, ctx: ASLParser.State_machineContext) -> Program:
        return self.visit(ctx.program_decl())

    def visitQuery_language_decl(
        self, ctx: ASLParser.Query_language_declContext
    ) -> QueryLanguage:
        query_language_mode_int = ctx.children[-1].getSymbol().type
        query_language_mode = QueryLanguageMode(value=query_language_mode_int)
        return QueryLanguage(query_language_mode=query_language_mode)

    def visitAssign_template_value_terminal_float(
        self, ctx: ASLParser.Assign_template_value_terminal_floatContext
    ) -> AssignTemplateValueTerminalLit:
        float_value = float(ctx.NUMBER().getText())
        return AssignTemplateValueTerminalLit(value=float_value)

    def visitAssign_template_value_terminal_int(
        self, ctx: ASLParser.Assign_template_value_terminal_intContext
    ) -> AssignTemplateValueTerminalLit:
        int_value = int(ctx.INT().getText())
        return AssignTemplateValueTerminalLit(value=int_value)

    def visitAssign_template_value_terminal_bool(
        self, ctx: ASLParser.Assign_template_value_terminal_boolContext
    ) -> AssignTemplateValueTerminalLit:
        bool_term_rule: int = ctx.children[0].getSymbol().type
        bool_value: bool = bool_term_rule == ASLLexer.TRUE
        return AssignTemplateValueTerminalLit(value=bool_value)

    def visitAssign_template_value_terminal_null(
        self, ctx: ASLParser.Assign_template_value_terminal_nullContext
    ) -> AssignTemplateValueTerminalLit:
        return AssignTemplateValueTerminalLit(value=None)

    def visitAssign_template_value_terminal_string_jsonata(
        self, ctx: ASLParser.Assign_template_value_terminal_string_jsonataContext
    ) -> AssignTemplateValueTerminal:
        # Return a JSONata expression resolver or a suppressed depending on the current language mode.
        current_query_language = self._get_current_query_language()
        if current_query_language.query_language_mode == QueryLanguageMode.JSONata:
            string_jsonata: StringJSONata = self.visitString_jsonata(
                ctx.string_jsonata()
            )
            return AssignTemplateValueTerminalStringJSONata(
                string_jsonata=string_jsonata
            )
        else:
            inner_string_value = self._inner_string_of(
                parser_rule_context=ctx.string_jsonata()
            )
            return AssignTemplateValueTerminalLit(value=inner_string_value)

    def visitAssign_template_value_terminal_string_literal(
        self, ctx: ASLParser.Assign_template_value_terminal_string_literalContext
    ) -> AssignTemplateValueTerminal:
        string_literal = self._inner_string_of(ctx.string_literal())
        return AssignTemplateValueTerminalLit(value=string_literal)

    def visitAssign_template_value(self, ctx: ASLParser.Assign_template_valueContext):
        return self.visit(ctx.children[0])

    def visitAssign_template_value_array(
        self, ctx: ASLParser.Assign_template_value_arrayContext
    ) -> AssignTemplateValueArray:
        values: list[AssignTemplateValue] = []
        for child in ctx.children:
            cmp: Optional[Component] = self.visit(child)
            if isinstance(cmp, AssignTemplateValue):
                values.append(cmp)
        return AssignTemplateValueArray(values=values)

    def visitAssign_template_value_object(
        self, ctx: ASLParser.Assign_template_value_objectContext
    ) -> AssignTemplateValueObject:
        bindings: list[AssignTemplateBinding] = []
        for child in ctx.children:
            cmp: Optional[Component] = self.visit(child)
            if isinstance(cmp, AssignTemplateBinding):
                bindings.append(cmp)
        return AssignTemplateValueObject(bindings=bindings)

    def visitAssign_template_binding_value(
        self, ctx: ASLParser.Assign_template_binding_valueContext
    ) -> AssignTemplateBindingValue:
        string_literal: StringLiteral = self.visitString_literal(
            ctx=ctx.string_literal()
        )
        assign_value: AssignTemplateValue = self.visit(ctx.assign_template_value())
        return AssignTemplateBindingValue(
            identifier=string_literal.literal_value, assign_value=assign_value
        )

    def visitAssign_template_binding_string_expression_simple(
        self, ctx: ASLParser.Assign_template_binding_string_expression_simpleContext
    ) -> AssignTemplateBindingStringExpressionSimple:
        identifier: str = self._inner_string_of(ctx.STRINGDOLLAR())
        identifier = identifier[:-2]
        string_expression_simple: StringExpressionSimple = (
            self.visitString_expression_simple(ctx.string_expression_simple())
        )
        return AssignTemplateBindingStringExpressionSimple(
            identifier=identifier, string_expression_simple=string_expression_simple
        )

    def visitAssign_decl_binding(
        self, ctx: ASLParser.Assign_decl_bindingContext
    ) -> AssignDeclBinding:
        binding: AssignTemplateBinding = self.visit(ctx.assign_template_binding())
        return AssignDeclBinding(binding=binding)

    def visitAssign_decl_body(
        self, ctx: ASLParser.Assign_decl_bodyContext
    ) -> list[AssignDeclBinding]:
        bindings: list[AssignDeclBinding] = []
        for child in ctx.children:
            cmp: Optional[Component] = self.visit(child)
            if isinstance(cmp, AssignDeclBinding):
                bindings.append(cmp)
        return bindings

    def visitAssign_decl(self, ctx: ASLParser.Assign_declContext) -> AssignDecl:
        declaration_bindings: list[AssignDeclBinding] = self.visit(
            ctx.assign_decl_body()
        )
        return AssignDecl(declaration_bindings=declaration_bindings)

    def visitJsonata_template_value_terminal_float(
        self, ctx: ASLParser.Jsonata_template_value_terminal_floatContext
    ) -> JSONataTemplateValueTerminalLit:
        float_value = float(ctx.NUMBER().getText())
        return JSONataTemplateValueTerminalLit(value=float_value)

    def visitJsonata_template_value_terminal_int(
        self, ctx: ASLParser.Jsonata_template_value_terminal_intContext
    ) -> JSONataTemplateValueTerminalLit:
        int_value = int(ctx.INT().getText())
        return JSONataTemplateValueTerminalLit(value=int_value)

    def visitJsonata_template_value_terminal_bool(
        self, ctx: ASLParser.Jsonata_template_value_terminal_boolContext
    ) -> JSONataTemplateValueTerminalLit:
        bool_term_rule: int = ctx.children[0].getSymbol().type
        bool_value: bool = bool_term_rule == ASLLexer.TRUE
        return JSONataTemplateValueTerminalLit(value=bool_value)

    def visitJsonata_template_value_terminal_null(
        self, ctx: ASLParser.Jsonata_template_value_terminal_nullContext
    ) -> JSONataTemplateValueTerminalLit:
        return JSONataTemplateValueTerminalLit(value=None)

    def visitJsonata_template_value_terminal_string_jsonata(
        self, ctx: ASLParser.Jsonata_template_value_terminal_string_jsonataContext
    ) -> JSONataTemplateValueTerminalStringJSONata:
        string_jsonata: StringJSONata = self.visitString_jsonata(ctx.string_jsonata())
        return JSONataTemplateValueTerminalStringJSONata(string_jsonata=string_jsonata)

    def visitJsonata_template_value_terminal_string_literal(
        self, ctx: ASLParser.Jsonata_template_value_terminal_string_literalContext
    ) -> JSONataTemplateValueTerminalLit:
        string = from_string_literal(ctx.string_literal())
        return JSONataTemplateValueTerminalLit(value=string)

    def visitJsonata_template_value(
        self, ctx: ASLParser.Jsonata_template_valueContext
    ) -> JSONataTemplateValue:
        return self.visit(ctx.children[0])

    def visitJsonata_template_value_array(
        self, ctx: ASLParser.Jsonata_template_value_arrayContext
    ) -> JSONataTemplateValueArray:
        values: list[JSONataTemplateValue] = []
        for child in ctx.children:
            cmp: Optional[Component] = self.visit(child)
            if isinstance(cmp, JSONataTemplateValue):
                values.append(cmp)
        return JSONataTemplateValueArray(values=values)

    def visitJsonata_template_value_object(
        self, ctx: ASLParser.Jsonata_template_value_objectContext
    ) -> JSONataTemplateValueObject:
        bindings: list[JSONataTemplateBinding] = []
        for child in ctx.children:
            cmp: Optional[Component] = self.visit(child)
            if isinstance(cmp, JSONataTemplateBinding):
                bindings.append(cmp)
        return JSONataTemplateValueObject(bindings=bindings)

    def visitJsonata_template_binding(
        self, ctx: ASLParser.Jsonata_template_bindingContext
    ) -> JSONataTemplateBinding:
        identifier: str = self._inner_string_of(ctx.string_literal())
        value: JSONataTemplateValue = self.visit(ctx.jsonata_template_value())
        return JSONataTemplateBinding(identifier=identifier, value=value)

    def visitArguments_string_jsonata(
        self, ctx: ASLParser.Arguments_string_jsonataContext
    ) -> ArgumentsStringJSONata:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONata, ctx=ctx
        )
        string_jsonata: StringJSONata = self.visitString_jsonata(ctx.string_jsonata())
        return ArgumentsStringJSONata(string_jsonata=string_jsonata)

    def visitArguments_jsonata_template_value_object(
        self, ctx: ASLParser.Arguments_jsonata_template_value_objectContext
    ) -> ArgumentsJSONataTemplateValueObject:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONata, ctx=ctx
        )
        jsonata_template_value_object: JSONataTemplateValueObject = self.visit(
            ctx.jsonata_template_value_object()
        )
        return ArgumentsJSONataTemplateValueObject(
            jsonata_template_value_object=jsonata_template_value_object
        )

    def visitOutput_decl(self, ctx: ASLParser.Output_declContext) -> Output:
        jsonata_template_value: JSONataTemplateValue = self.visit(
            ctx.jsonata_template_value()
        )
        return Output(jsonata_template_value=jsonata_template_value)

    def visitItems_array(self, ctx: ASLParser.Items_arrayContext) -> ItemsArray:
        jsonata_template_value_array: JSONataTemplateValueArray = self.visit(
            ctx.jsonata_template_value_array()
        )
        return ItemsArray(jsonata_template_value_array=jsonata_template_value_array)

    def visitItems_jsonata(self, ctx: ASLParser.Items_jsonataContext) -> ItemsJSONata:
        string_jsonata: StringJSONata = self.visitString_jsonata(ctx.string_jsonata())
        return ItemsJSONata(string_jsonata=string_jsonata)

    def visitString_sampler(
        self, ctx: ASLParser.String_samplerContext
    ) -> StringSampler:
        return self.visit(ctx.children[0])

    def visitString_literal(
        self, ctx: ASLParser.String_literalContext
    ) -> StringLiteral:
        string_literal = from_string_literal(parser_rule_context=ctx)
        return StringLiteral(literal_value=string_literal)

    def visitString_jsonpath(
        self, ctx: ASLParser.String_jsonpathContext
    ) -> StringJsonPath:
        json_path: str = self._inner_string_of(parser_rule_context=ctx)
        return StringJsonPath(json_path=json_path)

    def visitString_context_path(
        self, ctx: ASLParser.String_context_pathContext
    ) -> StringContextPath:
        context_object_path: str = self._inner_string_of(parser_rule_context=ctx)
        return StringContextPath(context_object_path=context_object_path)

    def visitString_variable_sample(
        self, ctx: ASLParser.String_variable_sampleContext
    ) -> StringVariableSample:
        query_language_mode: QueryLanguageMode = (
            self._get_current_query_language().query_language_mode
        )
        expression: str = self._inner_string_of(parser_rule_context=ctx)
        return StringVariableSample(
            query_language_mode=query_language_mode, expression=expression
        )

    def visitString_jsonata(
        self, ctx: ASLParser.String_jsonataContext
    ) -> StringJSONata:
        self._raise_if_query_language_is_not(
            query_language_mode=QueryLanguageMode.JSONata, ctx=ctx
        )
        expression = self._inner_jsonata_expr(ctx=ctx)
        return StringJSONata(expression=expression)

    def visitString_intrinsic_function(
        self, ctx: ASLParser.String_intrinsic_functionContext
    ) -> StringIntrinsicFunction:
        intrinsic_function_derivation = ctx.STRINGINTRINSICFUNC().getText()[1:-1]
        function, _ = IntrinsicParser.parse(intrinsic_function_derivation)
        return StringIntrinsicFunction(
            intrinsic_function_derivation=intrinsic_function_derivation,
            function=function,
        )



from localstack.packages import Package, package


@package(name="jpype-jsonata")
def jpype_jsonata_package() -> Package:
    """The Java-based jsonata library uses JPype and depends on a JVM installation."""
    from moto.stepfunctions.parser.packages import jpype_jsonata_package

    return jpype_jsonata_package



"""Handles incoming dsql requests, invokes methods, returns responses."""

import json
from urllib.parse import unquote

from moto.core.responses import ActionResult, BaseResponse

from .models import AuroraDSQLBackend, dsql_backends


class AuroraDSQLResponse(BaseResponse):
    """Handler for AuroraDSQL requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="dsql")

    @property
    def dsql_backend(self) -> AuroraDSQLBackend:
        """Return backend instance specific for this region."""
        return dsql_backends[self.current_account][self.region]

    def create_cluster(self) -> ActionResult:
        params = json.loads(self.body)
        deletion_protection_enabled = params.get("deletionProtectionEnabled", True)
        tags = params.get("tags")
        client_token = params.get("clientToken")
        cluster = self.dsql_backend.create_cluster(
            deletion_protection_enabled=deletion_protection_enabled,
            tags=tags,
            client_token=client_token,
        )
        return ActionResult(cluster)

    def delete_cluster(self) -> ActionResult:
        identifier = self.path.split("/")[-1]
        cluster = self.dsql_backend.delete_cluster(identifier=identifier)
        result = {
            "identifier": cluster.identifier,
            "arn": cluster.arn,
            "status": "DELETING",
            "creationTime": cluster.creation_time,
        }
        return ActionResult(result)

    def get_cluster(self) -> ActionResult:
        identifier = self.path.split("/")[-1]
        cluster = self.dsql_backend.get_cluster(identifier=identifier)
        return ActionResult(cluster)

    def get_vpc_endpoint_service_name(self) -> ActionResult:
        identifier = self.path.split("/")[-2]
        result = self.dsql_backend.get_vpc_endpoint_service_name(identifier)
        return ActionResult(result)

    def list_tags_for_resource(self) -> ActionResult:
        arn = unquote(self.path.split("/")[-1])
        identifier = arn.split("/")[-1]
        tags = self.dsql_backend.list_tags_for_resource(identifier)
        return ActionResult({"tags": tags})



"""account base URL and path."""

from .responses import AccountResponse

url_bases = [
    r"https?://account\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/putAlternateContact$": AccountResponse.dispatch,
    "{0}/getAlternateContact$": AccountResponse.dispatch,
    "{0}/deleteAlternateContact$": AccountResponse.dispatch,
}



import string
from email.utils import parseaddr
from typing import TYPE_CHECKING, Optional, Union

from moto.moto_api._internal import mock_random as random

if TYPE_CHECKING:
    from sesv2.models import SESV2Backend

    from .models import SESBackend


def random_hex(length: int) -> str:
    return "".join(random.choice(string.ascii_lowercase) for x in range(length))


def get_random_message_id() -> str:
    return f"{random_hex(16)}-{random_hex(8)}-{random_hex(4)}-{random_hex(4)}-{random_hex(4)}-{random_hex(12)}-{random_hex(6)}"


def is_valid_address(addr: str) -> Optional[str]:
    _, address = parseaddr(addr)
    address_parts = address.split("@")
    if len(address_parts) != 2 or not address_parts[1]:
        return "Missing domain"
    return None


def get_arn(
    backend: Union["SESBackend", "SESV2Backend"],
    service: str,
    resource: str,
) -> str:
    return f"arn:{backend.partition}:{service}:{backend.region_name}:{backend.account_id}:{resource}"



import json

from moto.core.responses import BaseResponse

from .models import MacieBackend, macie2_backends


class MacieResponse(BaseResponse):
    """Handles API requests for the Macie service."""

    def __init__(self) -> None:
        super().__init__(service_name="macie2")

    @property
    def macie_backend(self) -> MacieBackend:
        return macie2_backends[self.current_account][self.region]

    def create_invitations(self) -> str:
        account_ids = self._get_param("accountIds", [])
        self.macie_backend.create_invitations(account_ids=account_ids)
        return json.dumps({"unprocessedAccounts": []})

    def list_invitations(self) -> str:
        invitations = self.macie_backend.list_invitations()
        return json.dumps({"invitations": [inv.to_json() for inv in invitations]})

    def decline_invitations(self) -> str:
        account_ids = self._get_param("accountIds", [])
        self.macie_backend.decline_invitations(account_ids=account_ids)
        return json.dumps({"unprocessedAccounts": []})

    def accept_invitation(self) -> str:
        administrator_account_id = self._get_param("administratorAccountId")
        invitation_id = self._get_param("invitationId")
        self.macie_backend.accept_invitation(administrator_account_id, invitation_id)
        return "{}"

    def list_members(self) -> str:
        members = self.macie_backend.list_members()
        return json.dumps({"members": [m.to_json() for m in members]})

    def get_administrator_account(self) -> str:
        admin = self.macie_backend.get_administrator_account()
        if admin:
            response = {
                "administrator": {
                    "accountId": admin.administrator_account_id,
                    "relationshipStatus": "Enabled",
                }
            }
        else:
            # When there's no administrator, return an empty dict
            response = {}
        return json.dumps(response)

    def delete_member(self) -> str:
        member_account_id = self.path.split("/")[-1]
        self.macie_backend.delete_member(member_account_id)
        return "{}"

    def disassociate_member(self) -> str:
        member_account_id = self.path.split("/")[-1]
        self.macie_backend.disassociate_member(member_account_id)
        return "{}"

    def get_macie_session(self) -> str:
        session = self.macie_backend.get_macie_session()
        return json.dumps(session)

    def enable_macie(self) -> str:
        finding_publishing_frequency = self._get_param(
            "findingPublishingFrequency", "FIFTEEN_MINUTES"
        )
        status = self._get_param("status", "ENABLED")
        self.macie_backend.enable_macie(finding_publishing_frequency, status)
        return "{}"

    def disable_macie(self) -> str:
        self.macie_backend.disable_macie()
        return json.dumps({})

    def enable_organization_admin_account(self) -> str:
        admin_account_id = self._get_param("adminAccountId")
        self.macie_backend.enable_organization_admin_account(
            admin_account_id=admin_account_id
        )
        return json.dumps({})

    def list_organization_admin_accounts(self) -> str:
        admin_accounts = self.macie_backend.list_organization_admin_accounts()
        return json.dumps({"adminAccounts": admin_accounts})



from typing import Final

from moto.stepfunctions.parser.asl.component.component import Component


class CSVHeaders(Component):
    header_names: Final[list[str]]

    def __init__(self, header_names: list[str]):
        self.header_names = header_names



import copy

from moto.stepfunctions.parser.api import HistoryEventType, TaskFailedEventDetails
from moto.stepfunctions.parser.asl.component.common.error_name.custom_error_name import (
    CustomErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
    FailureEventException,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.mocking.mock_config import (
    MockedResponse,
    MockedResponseReturn,
    MockedResponseThrow,
)


def _eval_mocked_response_throw(
    env: Environment, mocked_response: MockedResponseThrow
) -> None:
    task_failed_event_details = TaskFailedEventDetails(
        error=mocked_response.error, cause=mocked_response.cause
    )
    error_name = CustomErrorName(mocked_response.error)
    failure_event = FailureEvent(
        env=env,
        error_name=error_name,
        event_type=HistoryEventType.TaskFailed,
        event_details=EventDetails(taskFailedEventDetails=task_failed_event_details),
    )
    raise FailureEventException(failure_event=failure_event)


def _eval_mocked_response_return(
    env: Environment, mocked_response: MockedResponseReturn
) -> None:
    payload_copy = copy.deepcopy(mocked_response.payload)
    env.stack.append(payload_copy)


def eval_mocked_response(env: Environment, mocked_response: MockedResponse) -> None:
    if isinstance(mocked_response, MockedResponseReturn):
        _eval_mocked_response_return(env=env, mocked_response=mocked_response)
    elif isinstance(mocked_response, MockedResponseThrow):
        _eval_mocked_response_throw(env=env, mocked_response=mocked_response)
    else:
        raise RuntimeError(f"Invalid MockedResponse type '{type(mocked_response)}'")



from moto.core.exceptions import ServiceException


class InvalidRequestException(ServiceException):
    code = "InvalidRequestException"


class ResourceNotFoundException(InvalidRequestException):
    error_code = "ResourceNotFound"


class InvalidCluster(InvalidRequestException):
    error_code = "NoSuchCluster"

    def __init__(self, cluster_id: str):
        message = f"Cluster id '{cluster_id}' is not valid."
        super().__init__(message)


class ValidationException(ServiceException):
    code = "ValidationException"



from moto.core.exceptions import ServiceException


class ElasticBeanstalkException(ServiceException):
    pass


class InvalidParameterValueError(ServiceException):
    code = "InvalidParameterValue"


class ResourceNotFoundException(ServiceException):
    code = "ResourceNotFoundException"



"""AccountBackend class with methods for supported APIs."""

from dataclasses import dataclass

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.utilities.utils import PARTITION_NAMES

from .exceptions import UnspecifiedContactType

ALLOWED_CONTACT_TYPES = ["SECURITY", "OPERATIONS", "BILLING"]


@dataclass
class AlternateContact(BaseModel):
    alternate_contact_type: str
    title: str
    name: str
    email_address: str
    phone_number: str


class AccountBackend(BaseBackend):
    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self._alternate_contacts: dict[str, AlternateContact] = {}

    def put_alternate_contact(
        self,
        alternate_contact_type: str,
        email_address: str,
        name: str,
        phone_number: str,
        title: str,
    ) -> None:
        self._alternate_contacts[alternate_contact_type] = AlternateContact(
            alternate_contact_type=alternate_contact_type,
            name=name,
            title=title,
            email_address=email_address,
            phone_number=phone_number,
        )

    def get_alternate_contact(self, alternate_contact_type: str) -> AlternateContact:
        if alternate_contact_type not in self._alternate_contacts:
            raise UnspecifiedContactType
        return self._alternate_contacts[alternate_contact_type]

    def delete_alternate_contact(self, alternate_contact_type: str) -> None:
        self._alternate_contacts.pop(alternate_contact_type, None)


account_backends = BackendDict(
    AccountBackend,
    "account",
    use_boto3_regions=False,
    additional_regions=PARTITION_NAMES,
)



"""ConnectCampaignServiceBackend class with methods for supported APIs."""

import uuid
from typing import Any, Optional
from urllib.parse import unquote

from moto.core import DEFAULT_ACCOUNT_ID
from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.utilities.paginator import paginate
from moto.utilities.tagging_service import TaggingService

from .exceptions import ResourceNotFoundException, ValidationException

PAGINATION_MODEL = {
    "list_campaigns": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "output_token": "nextToken",
        "unique_attribute": "id",
    }
}


class ConnectCampaign(BaseModel):
    def __init__(
        self,
        name: str,
        connect_instance_id: str,
        dialer_config: dict[str, Any],
        outbound_call_config: dict[str, Any],
        region: str,
        tags: Optional[dict[str, str]] = None,
    ) -> None:
        self.id = str(uuid.uuid4())
        self.name = name
        self.connect_instance_id = connect_instance_id
        self.state = "Initialized"
        self.dialer_config = dialer_config
        self.outbound_call_config = outbound_call_config
        self.region = region
        self.tags = tags or {}
        self.arn = f"arn:aws:connectcampaigns:{self.region}:{DEFAULT_ACCOUNT_ID}:campaign/{self.id}"

    def to_dict(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "arn": self.arn,
            "name": self.name,
            "connectInstanceId": self.connect_instance_id,
            "dialerConfig": self.dialer_config,
            "outboundCallConfig": self.outbound_call_config,
            "tags": self.tags,
        }


class ConnectInstanceConfig(BaseModel):
    def __init__(
        self, connect_instance_id: str, region: str, encryption_enabled: bool = False
    ) -> None:
        self.connect_instance_id = connect_instance_id
        self.region = region
        self.service_linked_role_arn = f"arn:aws:iam::{DEFAULT_ACCOUNT_ID}:role/aws-service-role/connectcampaigns.amazonaws.com/AWSServiceRoleForConnectCampaigns"

        self.encryption_config = {
            "enabled": encryption_enabled,
            "encryptionType": "KMS",
        }

        if encryption_enabled:
            self.encryption_config["keyArn"] = (
                f"arn:aws:kms:{region}:{DEFAULT_ACCOUNT_ID}:key/1234abcd-12ab-34cd-56ef-1234567890ab"
            )

    def to_dict(self) -> dict[str, Any]:
        return {
            "connectInstanceId": self.connect_instance_id,
            "serviceLinkedRoleArn": self.service_linked_role_arn,
            "encryptionConfig": self.encryption_config,
        }


class ConnectInstanceOnboardingJobStatus(BaseModel):
    def __init__(
        self,
        connect_instance_id: str,
        status: str = "SUCCEEDED",
        failure_code: Optional[str] = None,
    ) -> None:
        self.connect_instance_id = connect_instance_id
        self.status = status
        self.failure_code = failure_code

    def to_dict(self) -> dict[str, Any]:
        result = {"connectInstanceId": self.connect_instance_id, "status": self.status}

        if self.failure_code and self.status == "FAILED":
            result["failureCode"] = self.failure_code

        return result


class ConnectCampaignServiceBackend(BaseBackend):
    def __init__(self, region_name: str, account_id: str) -> None:
        super().__init__(region_name, account_id)
        self.campaigns: dict[str, ConnectCampaign] = {}
        self.instance_configs: dict[str, ConnectInstanceConfig] = {}
        self.onboarding_jobs: dict[str, ConnectInstanceOnboardingJobStatus] = {}
        self.tagger = TaggingService()

    def create_campaign(
        self,
        name: str,
        connect_instance_id: str,
        dialer_config: dict[str, Any],
        outbound_call_config: dict[str, Any],
        tags: Optional[dict[str, str]],
    ) -> tuple[str, str, dict[str, str]]:
        campaign = ConnectCampaign(
            name=name,
            connect_instance_id=connect_instance_id,
            dialer_config=dialer_config,
            outbound_call_config=outbound_call_config,
            region=self.region_name,
            tags=tags,
        )
        self.campaigns[campaign.id] = campaign
        return campaign.id, campaign.arn, campaign.tags

    def delete_campaign(self, id: str) -> None:
        del self.campaigns[id]

    def describe_campaign(self, id: str) -> dict[str, Any]:
        if id not in self.campaigns:
            raise ResourceNotFoundException(f"Campaign with id {id} not found")

        campaign = self.campaigns[id]
        return campaign.to_dict()

    def get_connect_instance_config(self, connect_instance_id: str) -> dict[str, Any]:
        if not connect_instance_id:
            raise ValidationException("connectInstanceId is a required parameter")

        if connect_instance_id == "invalid-id":
            raise ResourceNotFoundException(
                f"Connect instance with id {connect_instance_id} not found"
            )

        if connect_instance_id in self.instance_configs:
            return self.instance_configs[connect_instance_id].to_dict()

        instance_config = ConnectInstanceConfig(
            connect_instance_id=connect_instance_id, region=self.region_name
        )
        self.instance_configs[connect_instance_id] = instance_config

        return instance_config.to_dict()

    def start_instance_onboarding_job(
        self, connect_instance_id: str, encryption_config: dict[str, Any]
    ) -> dict[str, Any]:
        if not connect_instance_id:
            raise ValidationException("connectInstanceId is a required parameter")

        if connect_instance_id == "invalid-id":
            raise ResourceNotFoundException(
                f"Connect instance with id {connect_instance_id} not found"
            )

        if not encryption_config:
            raise ValidationException("encryptionConfig is a required parameter")

        if "enabled" not in encryption_config:
            raise ValidationException(
                "enabled is a required parameter in encryptionConfig"
            )

        if encryption_config.get("enabled") and "keyArn" not in encryption_config:
            raise ValidationException("keyArn is required when encryption is enabled")

        job_status = ConnectInstanceOnboardingJobStatus(
            connect_instance_id=connect_instance_id
        )

        self.onboarding_jobs[connect_instance_id] = job_status

        instance_config = ConnectInstanceConfig(
            connect_instance_id=connect_instance_id,
            region=self.region_name,
            encryption_enabled=encryption_config.get("enabled", False),
        )

        if encryption_config.get("enabled") and "keyArn" in encryption_config:
            instance_config.encryption_config["keyArn"] = encryption_config["keyArn"]

        self.instance_configs[connect_instance_id] = instance_config

        return job_status.to_dict()

    def start_campaign(self, id: str) -> None:
        if id not in self.campaigns:
            raise ResourceNotFoundException(f"Campaign with id {id} not found")
        self.campaigns[id].state = "Running"
        return

    def stop_campaign(self, id: str) -> None:
        if id not in self.campaigns:
            raise ResourceNotFoundException(f"Campaign with id {id} not found")
        self.campaigns[id].state = "Stopped"
        return

    def pause_campaign(self, id: str) -> None:
        if id not in self.campaigns:
            raise ResourceNotFoundException(f"Campaign with id {id} not found")
        self.campaigns[id].state = "Paused"
        return

    def resume_campaign(self, id: str) -> None:
        if id not in self.campaigns:
            raise ResourceNotFoundException(f"Campaign with id {id} not found")
        self.campaigns[id].state = "Running"
        return

    def get_campaign_state(self, id: str) -> str:
        if id not in self.campaigns:
            raise ResourceNotFoundException(f"Campaign with id {id} not found")
        return self.campaigns[id].state

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_campaigns(
        self,
        filters: Optional[dict[str, dict[str, str]]],
        max_results: Optional[int],
        next_token: Optional[str],
    ) -> list[dict[str, str]]:
        filtered_campaigns = list(self.campaigns.values())

        if filters and "instanceIdFilter" in filters:
            filter_value = filters["instanceIdFilter"]["value"]
            filter_operator = filters["instanceIdFilter"]["operator"]

            operator_logic = {
                "Eq": lambda campaign: campaign.connect_instance_id == filter_value,
                "Ne": lambda campaign: campaign.connect_instance_id != filter_value,
                "Contains": lambda campaign: filter_value
                in campaign.connect_instance_id,
                "StartsWith": lambda campaign: campaign.connect_instance_id.startswith(
                    filter_value
                ),
            }

            if filter_operator not in operator_logic:
                raise ValidationException(f"Unsupported operator: {filter_operator}")

            filtered_campaigns = list(
                filter(operator_logic[filter_operator], filtered_campaigns)
            )

        campaign_summary_list = [
            {
                "id": campaign.id,
                "arn": campaign.arn,
                "name": campaign.name,
                "connectInstanceId": campaign.connect_instance_id,
            }
            for campaign in filtered_campaigns
        ]
        return campaign_summary_list

    def tag_resource(self, arn: str, tags: dict[str, str]) -> None:
        arn = unquote(arn)
        campaign = None
        for c in self.campaigns.values():
            if c.arn == arn:
                campaign = c
                break

        if campaign is None:
            raise ResourceNotFoundException(f"Resource {arn} not found")

        tag_list = [{"Key": k, "Value": v} for k, v in tags.items()]
        campaign.tags.update(tags)
        self.tagger.tag_resource(arn, tag_list)
        return

    def untag_resource(self, arn: str, tag_keys: list[str]) -> None:
        arn = unquote(arn)
        campaign = None
        for c in self.campaigns.values():
            if c.arn == arn:
                campaign = c
                break

        if campaign is None:
            raise ResourceNotFoundException(f"Resource {arn} not found")

        if not isinstance(tag_keys, list):
            tag_keys = [tag_keys]
        if not tag_keys:
            raise ValidationException("tagKeys is a required parameter")

        for tag in tag_keys:
            campaign.tags.pop(tag, None)
        self.tagger.untag_resource_using_names(arn, tag_keys)
        return

    def list_tags_for_resource(self, arn: str) -> dict[str, str]:
        arn = unquote(arn)
        campaign = None
        for c in self.campaigns.values():
            if c.arn == arn:
                campaign = c
                break

        if campaign is None:
            raise ResourceNotFoundException(f"Resource {arn} not found")

        return self.tagger.get_tag_dict_for_resource(arn)


connectcampaigns_backends = BackendDict(
    ConnectCampaignServiceBackend, "connectcampaigns"
)



from __future__ import annotations

import random
import string
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from moto.dms.models import FakeReplicationTask


def random_id(uppercase: bool = True, length: int = 13) -> str:
    ascii_set = string.ascii_uppercase if uppercase else string.ascii_lowercase
    chars = list(range(10)) + list(ascii_set)
    resource_id = random.choice(ascii_set) + "".join(
        str(random.choice(chars)) for _ in range(length - 1)
    )
    return resource_id


def match_task_arn(task: FakeReplicationTask, arns: list[str]) -> bool:
    return task.arn in arns


def match_task_id(task: FakeReplicationTask, ids: list[str]) -> bool:
    return task.id in ids


def match_task_migration_type(
    task: FakeReplicationTask, migration_types: list[str]
) -> bool:
    return task.migration_type in migration_types


def match_task_endpoint_arn(
    task: FakeReplicationTask, endpoint_arns: list[str]
) -> bool:
    return (
        task.source_endpoint_arn in endpoint_arns
        or task.target_endpoint_arn in endpoint_arns
    )


def match_task_replication_instance_arn(
    task: FakeReplicationTask, replication_instance_arns: list[str]
) -> bool:
    return task.replication_instance_arn in replication_instance_arns


task_filter_functions = {
    "replication-task-arn": match_task_arn,
    "replication-task-id": match_task_id,
    "migration-type": match_task_migration_type,
    "endpoint-arn": match_task_endpoint_arn,
    "replication-instance-arn": match_task_replication_instance_arn,
}


def filter_tasks(
    tasks: list[FakeReplicationTask], filters: list[dict[str, Any]]
) -> Any:
    matching_tasks = tasks

    for f in filters:
        filter_function = task_filter_functions.get(f["Name"])

        if not filter_function:
            continue

        matching_tasks = list(
            filter(lambda task: filter_function(task, f["Values"]), matching_tasks)
        )

    return matching_tasks



"""pipes base URL and path."""

from .responses import EventBridgePipesResponse

url_bases = [
    r"https?://pipes\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/v1/pipes/(?P<Name>[^/]+)$": EventBridgePipesResponse.dispatch,
    "{0}/tags/(?P<resourceArn>.+)$": EventBridgePipesResponse.dispatch,
    "{0}/v1/pipes$": EventBridgePipesResponse.dispatch,
}



"""clouddirectory base URL and path."""

from .responses import CloudDirectoryResponse

url_bases = [
    r"https?://clouddirectory\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/amazonclouddirectory/2017-01-11/directory/create$": CloudDirectoryResponse.dispatch,
    "{0}/amazonclouddirectory/2017-01-11/schema/create$": CloudDirectoryResponse.dispatch,
    "{0}/amazonclouddirectory/2017-01-11/directory/list$": CloudDirectoryResponse.dispatch,
    "{0}/amazonclouddirectory/2017-01-11/tags/add$": CloudDirectoryResponse.dispatch,
    "{0}/amazonclouddirectory/2017-01-11/tags/remove$": CloudDirectoryResponse.dispatch,
    "{0}/amazonclouddirectory/2017-01-11/directory$": CloudDirectoryResponse.dispatch,
    "{0}/amazonclouddirectory/2017-01-11/directory/get$": CloudDirectoryResponse.dispatch,
    "{0}/amazonclouddirectory/2017-01-11/tags$": CloudDirectoryResponse.dispatch,
    "{0}/amazonclouddirectory/2017-01-11/schema$": CloudDirectoryResponse.dispatch,
    "{0}/amazonclouddirectory/2017-01-11/schema/development$": CloudDirectoryResponse.dispatch,
    "{0}/amazonclouddirectory/2017-01-11/schema/published$": CloudDirectoryResponse.dispatch,
    "{0}/amazonclouddirectory/2017-01-11/schema/apply$": CloudDirectoryResponse.dispatch,
    "{0}/amazonclouddirectory/2017-01-11/schema/publish$": CloudDirectoryResponse.dispatch,
}



from copy import deepcopy
from typing import Any

from moto.core.responses import ActionResult, EmptyResult
from moto.core.types import Base64EncodedString
from moto.core.utils import camelcase_to_underscores
from moto.ec2.exceptions import (
    InvalidParameterCombination,
    InvalidRequest,
    MissingParameterError,
)
from moto.ec2.utils import filter_iam_instance_profiles, parse_user_data

from ._base_response import EC2BaseResponse


class InstanceResponse(EC2BaseResponse):
    def describe_instances(self) -> ActionResult:
        self.error_on_dryrun()
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2/client/describe_instances.html
        # You cannot specify this(MaxResults) parameter and the instance IDs parameter in the same request.
        if "InstanceId.1" in self.data and "MaxResults" in self.data:
            raise InvalidParameterCombination(
                "The parameter instancesSet cannot be used with the parameter maxResults"
            )
        filter_dict = self._filters_from_querystring()
        instance_ids = self._get_param("InstanceIds", [])
        token = self._get_param("NextToken")
        if instance_ids:
            reservations = self.ec2_backend.get_reservations_by_instance_ids(
                instance_ids, filters=filter_dict
            )
        else:
            reservations = self.ec2_backend.describe_instances(filters=filter_dict)

        reservation_ids = [reservation.id for reservation in reservations]
        if token:
            start = reservation_ids.index(token) + 1
        else:
            start = 0
        max_results = int(self._get_param("MaxResults", 100))
        reservations_resp = reservations[start : start + max_results]
        next_token = None
        if max_results and len(reservations) > (start + max_results):
            next_token = reservations_resp[-1].id
        result = {"Reservations": reservations_resp, "NextToken": next_token}
        return ActionResult(result)

    def run_instances(self) -> ActionResult:
        min_count = int(self._get_param("MinCount", if_none="1"))
        image_id = self._get_param("ImageId")
        user_data = parse_user_data(self._get_param("UserData"))
        security_group_names = self._get_param("SecurityGroups", [])
        kwargs = {
            "instance_type": self._get_param("InstanceType", "m1.small"),
            "is_instance_type_default": not self._get_param("InstanceType"),
            "placement": self._get_param("Placement.AvailabilityZone"),
            "placement_hostid": self._get_param("Placement.HostId"),
            "region_name": self.region,
            "subnet_id": self._get_param("SubnetId"),
            "key_name": self._get_param("KeyName"),
            "security_group_ids": self._get_param("SecurityGroupIds", []),
            "nics": self._get_param("NetworkInterfaces", []),
            "private_ip": self._get_param("PrivateIpAddress"),
            "associate_public_ip": self._get_param("AssociatePublicIpAddress"),
            "tags": self._parse_tag_specification(),
            "ebs_optimized": self._get_param("EbsOptimized") or False,
            "disable_api_stop": self._get_param("DisableApiStop") or False,
            "instance_market_options": self._get_param(
                "InstanceMarketOptions.MarketType", {}
            ),
            "instance_initiated_shutdown_behavior": self._get_param(
                "InstanceInitiatedShutdownBehavior"
            ),
            "launch_template": self._get_param("LaunchTemplate", {}),
            "hibernation_options": self._get_param("HibernationOptions", {}),
            "iam_instance_profile_name": self._get_param("IamInstanceProfile.Name"),
            "iam_instance_profile_arn": self._get_param("IamInstanceProfile.Arn"),
            "monitoring_state": "enabled"
            if self._get_param("Monitoring.Enabled")
            else "disabled",
            "ipv6_address_count": self._get_int_param("Ipv6AddressCount"),
            "metadata_options": self._get_param("MetadataOptions", {}),
            "client_token": self._get_param("ClientToken", "ABCDE"),
        }
        if len(kwargs["nics"]) and kwargs["subnet_id"]:
            raise InvalidParameterCombination(
                msg="Network interfaces and an instance-level subnet ID may not be specified on the same request"
            )

        mappings = self._parse_block_device_mapping()
        if mappings:
            kwargs["block_device_mappings"] = mappings

        iam_instance_profile_name = kwargs.get("iam_instance_profile_name")
        iam_instance_profile_arn = kwargs.get("iam_instance_profile_arn")
        if iam_instance_profile_arn or iam_instance_profile_name:
            # Validate the profile exists, before we error_on_dryrun and run_instances
            filter_iam_instance_profiles(
                self.current_account,
                partition=self.partition,
                iam_instance_profile_arn=iam_instance_profile_arn,
                iam_instance_profile_name=iam_instance_profile_name,
            )
        self.error_on_dryrun()
        new_reservation = self.ec2_backend.run_instances(
            image_id, min_count, user_data, security_group_names, **kwargs
        )
        if iam_instance_profile_name:
            self.ec2_backend.associate_iam_instance_profile(
                instance_id=new_reservation.instances[0].id,
                iam_instance_profile_name=iam_instance_profile_name,
            )
        if iam_instance_profile_arn:
            self.ec2_backend.associate_iam_instance_profile(
                instance_id=new_reservation.instances[0].id,
                iam_instance_profile_arn=iam_instance_profile_arn,
            )
        return ActionResult(new_reservation)

    def terminate_instances(self) -> ActionResult:
        instance_ids = self._get_param("InstanceIds", [])

        self.error_on_dryrun()

        instances = self.ec2_backend.terminate_instances(instance_ids)
        from moto.autoscaling import autoscaling_backends
        from moto.elbv2 import elbv2_backends

        autoscaling_backends[self.current_account][
            self.region
        ].notify_terminate_instances(instance_ids)
        elbv2_backends[self.current_account][self.region].notify_terminate_instances(
            instance_ids
        )
        result = {
            "TerminatingInstances": [
                {
                    "InstanceId": instance.id,
                    "CurrentState": {"Code": 32, "Name": "shutting-down"},
                    "PreviousState": previous_state,
                }
                for instance, previous_state in instances
            ]
        }
        return ActionResult(result)

    def reboot_instances(self) -> ActionResult:
        instance_ids = self._get_param("InstanceIds", [])
        self.error_on_dryrun()
        self.ec2_backend.reboot_instances(instance_ids)
        return EmptyResult()

    def stop_instances(self) -> ActionResult:
        instance_ids = self._get_param("InstanceIds", [])
        self.error_on_dryrun()
        instances = self.ec2_backend.stop_instances(instance_ids)
        result = {
            "StoppingInstances": [
                {
                    "InstanceId": instance.id,
                    "CurrentState": {"Code": 64, "Name": "stopping"},
                    "PreviousState": previous_state,
                }
                for instance, previous_state in instances
            ]
        }
        return ActionResult(result)

    def start_instances(self) -> ActionResult:
        instance_ids = self._get_param("InstanceIds", [])
        self.error_on_dryrun()
        instances = self.ec2_backend.start_instances(instance_ids)
        result = {
            "StartingInstances": [
                {
                    "InstanceId": instance.id,
                    "CurrentState": {"Code": 0, "Name": "pending"},
                    "PreviousState": previous_state,
                }
                for instance, previous_state in instances
            ]
        }
        return ActionResult(result)

    def describe_instance_status(self) -> ActionResult:
        instance_ids = self._get_param("InstanceIds", [])
        include_all_instances = self._get_bool_param("IncludeAllInstances", False)
        filters = [
            {"name": k, "values": v}
            for k, v in self._filters_from_querystring().items()
        ]
        instances = self.ec2_backend.describe_instance_status(
            instance_ids, include_all_instances, filters
        )
        result = {"InstanceStatuses": instances}
        return ActionResult(result)

    def describe_instance_types(self) -> ActionResult:
        instance_type_filters = self._get_param("InstanceTypes", [])
        filter_dict = self._filters_from_querystring()
        instance_types = self.ec2_backend.describe_instance_types(
            instance_type_filters, filter_dict
        )
        result = {"InstanceTypes": instance_types}
        return ActionResult(result)

    def describe_instance_type_offerings(self) -> ActionResult:
        location_type_filters = self._get_param("LocationType")
        filter_dict = self._filters_from_querystring()
        offerings = self.ec2_backend.describe_instance_type_offerings(
            location_type_filters, filter_dict
        )
        result = {"InstanceTypeOfferings": offerings}
        return ActionResult(result)

    def describe_instance_attribute(self) -> ActionResult:
        # TODO this and modify below should raise IncorrectInstanceState if
        # instance not in stopped state
        attribute = self._get_param("Attribute")
        instance_id = self._get_param("InstanceId")
        instance, value = self.ec2_backend.describe_instance_attribute(
            instance_id, attribute
        )
        attribute_name = attribute[0].upper() + attribute[1:]
        attribute_value: Any = {"Value": value}
        if attribute_name == "GroupSet":
            attribute_name = "Groups"
            attribute_value = [{"GroupId": group.id} for group in value]
        elif attribute_name == "UserData" and value is None:
            attribute_value = ""
        result = {"InstanceId": instance.id, attribute_name: attribute_value}
        return ActionResult(result)

    def describe_instance_credit_specifications(self) -> ActionResult:
        instance_ids = self._get_param("InstanceIds", [])
        instances = self.ec2_backend.describe_instance_credit_specifications(
            instance_ids
        )
        result = {
            "InstanceCreditSpecifications": [
                {"InstanceId": instance.id, "CpuCredits": "standard"}
                for instance in instances
            ],
        }
        return ActionResult(result)

    def modify_instance_attribute(self) -> ActionResult:
        handlers = [
            self._attribute_value_handler,
            self._dot_value_instance_attribute_handler,
            self._block_device_mapping_handler,
            self._security_grp_instance_attribute_handler,
        ]

        for handler in handlers:
            success = handler()
            if success:
                return EmptyResult()

        msg = (
            "This specific call to ModifyInstanceAttribute has not been"
            " implemented in Moto yet. Feel free to open an issue at"
            " https://github.com/getmoto/moto/issues"
        )
        raise NotImplementedError(msg)

    def modify_instance_metadata_options(self) -> ActionResult:
        instance_id = self._get_param("InstanceId")
        tokens = self._get_param("HttpTokens")
        hop_limit = self._get_int_param("HttpPutResponseHopLimit")
        endpoint = self._get_param("HttpEndpoint")
        dry_run = False
        http_protocol = self._get_param("HttpProtocolIpv6")
        metadata_tags = self._get_param("InstanceMetadataTags")
        options = self.ec2_backend.modify_instance_metadata_options(
            instance_id=instance_id,
            http_tokens=tokens,
            hop_limit=hop_limit,
            http_endpoint=endpoint,
            dry_run=dry_run,
            http_protocol=http_protocol,
            metadata_tags=metadata_tags,
        )
        result = {"InstanceId": instance_id, "InstanceMetadataOptions": options}
        return ActionResult(result)

    def get_instance_uefi_data(self) -> ActionResult:
        instance_id = self._get_param("InstanceId")
        uefi_data = self.ec2_backend.get_instance_uefi_data(instance_id)
        result = {
            "InstanceId": instance_id,
            "UefiData": Base64EncodedString.from_encoded_bytes(uefi_data),
        }
        return ActionResult(result)

    def _block_device_mapping_handler(self) -> bool:
        """
        Handles requests which are generated by code similar to:

            instance.modify_attribute(
                BlockDeviceMappings=[{
                    'DeviceName': '/dev/sda1',
                    'Ebs': {'DeleteOnTermination': True}
                }]
            )

        For now we only support the "BlockDeviceMapping.1.Ebs.DeleteOnTermination"
        configuration, but it should be trivial to add anything else.
        """
        attribute_modified = False
        for mapping in self._get_param("BlockDeviceMappings", []):
            device_name = mapping.get("DeviceName")
            del_on_term_value = mapping.get("Ebs", {}).get("DeleteOnTermination")

            instance_id = self._get_param("InstanceId")
            instance = self.ec2_backend.get_instance(instance_id)

            self.error_on_dryrun()

            if del_on_term_value is None:
                continue
            block_device_type = instance.block_device_mapping[device_name]
            block_device_type.delete_on_termination = del_on_term_value

            attribute_modified = True

        if attribute_modified:
            return True
        return False

    def _dot_value_instance_attribute_handler(self) -> bool:
        attribute_modified = False
        for attribute in (
            "InstanceType",
            "Kernel",
            "UserData",
            "DisableApiTermination",
            "SourceDestCheck",
            "EbsOptimized",
            "DisableApiStop",
        ):
            if self._get_param(f"{attribute}.Value") is not None:
                self.error_on_dryrun()
                instance_id = self._get_param("InstanceId")
                attr_name = camelcase_to_underscores(attribute)
                attr_value = self._get_param(f"{attribute}.Value")
                if attribute == "UserData" and attr_value:
                    attr_value = parse_user_data(attr_value)
                self.ec2_backend.modify_instance_attribute(
                    instance_id, attr_name, attr_value
                )
                attribute_modified = True
                break
        if attribute_modified:
            return True
        return False

    def _attribute_value_handler(self) -> bool:
        attribute_key = self._get_param("Attribute")

        if attribute_key is None:
            return False

        self.error_on_dryrun()

        value = self._get_param("Value")
        normalized_attribute = camelcase_to_underscores(attribute_key)
        instance_id = self._get_param("InstanceId")
        self.ec2_backend.modify_instance_attribute(
            instance_id, normalized_attribute, value
        )
        return True

    def _security_grp_instance_attribute_handler(self) -> bool:
        new_security_grp_list = self._get_param("Groups", [])
        self.error_on_dryrun()
        instance_id = self._get_param("InstanceId")
        self.ec2_backend.modify_instance_security_groups(
            instance_id, new_security_grp_list
        )
        return True

    def _parse_block_device_mapping(self) -> list[dict[str, Any]]:
        device_mappings = self._get_param("BlockDeviceMappings", [])
        mappings = []
        for device_mapping in device_mappings:
            self._validate_block_device_mapping(device_mapping)
            device_template: dict[str, Any] = deepcopy(BLOCK_DEVICE_MAPPING_TEMPLATE)
            device_template["VirtualName"] = device_mapping.get("VirtualName")
            device_template["DeviceName"] = device_mapping.get("DeviceName")
            device_template["Ebs"]["SnapshotId"] = device_mapping.get("Ebs", {}).get(
                "SnapshotId"
            )
            device_template["Ebs"]["VolumeSize"] = device_mapping.get("Ebs", {}).get(
                "VolumeSize"
            )
            device_template["Ebs"]["DeleteOnTermination"] = device_mapping.get(
                "Ebs", {}
            ).get("DeleteOnTermination", False)
            device_template["Ebs"]["VolumeType"] = device_mapping.get("Ebs", {}).get(
                "VolumeType"
            )
            device_template["Ebs"]["Iops"] = device_mapping.get("Ebs", {}).get("Iops")
            device_template["Ebs"]["Encrypted"] = device_mapping.get("Ebs", {}).get(
                "VolumeSize", False
            )
            device_template["Ebs"]["KmsKeyId"] = device_mapping.get("Ebs", {}).get(
                "KmsKeyId"
            )
            device_template["NoDevice"] = device_mapping.get("NoDevice")
            mappings.append(device_template)

        return mappings

    @staticmethod
    def _validate_block_device_mapping(device_mapping: dict[str, Any]) -> None:  # type: ignore[misc]
        from botocore import __version__ as botocore_version

        if "NoDevice" in device_mapping:
            assert isinstance(device_mapping["NoDevice"], str), (
                f"botocore {botocore_version} isn't limiting NoDevice to str type anymore, it is type:{type(device_mapping['no_device'])}"
            )
            if device_mapping["NoDevice"] == "":
                # the only legit value it can have is empty string
                # and none of the other checks here matter if NoDevice
                # is being used
                return
            else:
                raise InvalidRequest()

        if "Ebs" not in device_mapping:
            raise MissingParameterError("ebs")
        if not device_mapping.get("Ebs", {}).get(
            "VolumeSize"
        ) and not device_mapping.get("Ebs", {}).get("SnapshotId"):
            raise MissingParameterError("size or snapshotId")


BLOCK_DEVICE_MAPPING_TEMPLATE = {
    "VirtualName": None,
    "DeviceName": None,
    "NoDevice": None,
    "Ebs": {
        "SnapshotId": None,
        "VolumeSize": None,
        "DeleteOnTermination": None,
        "VolumeType": None,
        "Iops": None,
        "Encrypted": None,
    },
}



from moto.core.exceptions import ServiceException


class CloudWatchException(ServiceException):
    pass


class InvalidFormat(CloudWatchException):
    code = "InvalidFormat"


class InvalidParameterValue(CloudWatchException):
    code = "InvalidParameterValue"


class InvalidParameterCombination(CloudWatchException):
    code = "InvalidParameterCombination"


class ResourceNotFound(CloudWatchException):
    code = "ResourceNotFound"


class ResourceNotFoundException(CloudWatchException):
    code = "ResourceNotFoundException"
    message = "Unknown"


class ValidationError(CloudWatchException):
    code = "ValidationError"


class DashboardInvalidInputError(CloudWatchException):
    code = "InvalidParameterInput"



import abc
from typing import Final

from moto.stepfunctions.parser.asl.component.common.string.string_expression import (
    StringExpression,
    StringIntrinsicFunction,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.component.intrinsic.functionname.state_function_name_types import (
    StatesFunctionNameType,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment

_STRING_RETURN_FUNCTIONS: Final[set[str]] = {
    typ.name()
    for typ in [
        StatesFunctionNameType.Format,
        StatesFunctionNameType.JsonToString,
        StatesFunctionNameType.ArrayGetItem,
        StatesFunctionNameType.Base64Decode,
        StatesFunctionNameType.Base64Encode,
        StatesFunctionNameType.Hash,
        StatesFunctionNameType.UUID,
    ]
}


class CauseDecl(EvalComponent, abc.ABC): ...


class Cause(CauseDecl):
    string_expression: Final[StringExpression]

    def __init__(self, string_expression: StringExpression):
        self.string_expression = string_expression

    def _eval_body(self, env: Environment) -> None:
        self.string_expression.eval(env=env)


class CausePath(Cause):
    def __init__(self, string_expression: StringExpression):
        super().__init__(string_expression=string_expression)
        if isinstance(string_expression, StringIntrinsicFunction):
            if string_expression.function.name.name not in _STRING_RETURN_FUNCTIONS:
                raise ValueError(
                    f"Unsupported Intrinsic Function for CausePath declaration: '{string_expression.intrinsic_function_derivation}'."
                )



from moto.core.responses import ActionResult, EmptyResult
from moto.ec2.models import validate_resource_ids

from ._base_response import EC2BaseResponse


class TagResponse(EC2BaseResponse):
    def create_tags(self) -> ActionResult:
        resource_ids = self._get_param("Resources", [])
        validate_resource_ids(resource_ids)
        self.ec2_backend.do_resources_exist(resource_ids)
        tags = {tag["Key"]: tag["Value"] for tag in self._get_param("Tags", [])}

        self.error_on_dryrun()

        self.ec2_backend.create_tags(resource_ids, tags)
        return EmptyResult()

    def delete_tags(self) -> ActionResult:
        resource_ids = self._get_param("Resources", [])
        validate_resource_ids(resource_ids)
        tags = {tag["Key"]: tag.get("Value") for tag in self._get_param("Tags", [])}

        self.error_on_dryrun()

        self.ec2_backend.delete_tags(resource_ids, tags)
        return EmptyResult()

    def describe_tags(self) -> ActionResult:
        filters = self._filters_from_querystring()

        self.error_on_dryrun()

        tags = self.ec2_backend.describe_tags(filters=filters)
        result = {"Tags": tags}
        return ActionResult(result)



from enum import Enum
from typing import Any, Optional, Union

from moto.dynamodb.exceptions import KeyIsEmptyStringException, MockValidationException
from moto.utilities.tokenizer import GenericTokenizer


class EXPRESSION_STAGES(Enum):
    INITIAL_STAGE = "INITIAL_STAGE"  # Can be a hash key, range key, or function
    KEY_NAME = "KEY_NAME"
    KEY_VALUE = "KEY_VALUE"
    COMPARISON = "COMPARISON"
    EOF = "EOF"


def get_keys(schema: list[dict[str, str]], key_type: str) -> list[str]:
    """Get all keys of the specified type in schema order."""
    return [key["AttributeName"] for key in schema if key["KeyType"] == key_type]


def parse_expression(
    key_condition_expression: str,
    expression_attribute_values: dict[str, dict[str, str]],
    expression_attribute_names: dict[str, str],
    schema: list[dict[str, str]],
) -> tuple[
    list[tuple[str, dict[str, Any]]],
    list[tuple[str, str, list[dict[str, Any]]]],
    list[str],
]:
    """
    Parse a KeyConditionExpression using the provided expression attribute names/values

    key_condition_expression:    hashkey = :id AND :sk = val
    expression_attribute_names:  {":sk": "sortkey"}
    expression_attribute_values: {":id": {"S": "some hash key"}}
    schema:                      [{'AttributeName': 'hashkey', 'KeyType': 'HASH'}, {"AttributeName": "sortkey", "KeyType": "RANGE"}]

    Returns:
        - hash_key_conditions: List of (attr_name, value) for all hash keys
        - range_key_conditions: List of (attr_name, comparison, values) for range keys
        - expression_attribute_names_used: List of expression attribute names used
    """

    current_stage: Optional[EXPRESSION_STAGES] = None
    current_phrase = ""
    key_name = comparison = ""
    key_values: list[Union[dict[str, str], str]] = []
    expression_attribute_names_used: list[str] = []
    results: list[tuple[str, str, Any]] = []
    tokenizer = GenericTokenizer(key_condition_expression)
    for crnt_char in tokenizer:
        if crnt_char == " ":
            if current_stage == EXPRESSION_STAGES.INITIAL_STAGE:
                tokenizer.skip_white_space()
                if tokenizer.peek() == "(":
                    # begins_with(sk, :sk) and primary = :pk
                    #            ^
                    continue
                else:
                    # start_date < :sk and primary = :pk
                    #            ^
                    if expression_attribute_names.get(current_phrase):
                        key_name = expression_attribute_names[current_phrase]
                        expression_attribute_names_used.append(current_phrase)
                    else:
                        key_name = current_phrase
                    current_phrase = ""
                    current_stage = EXPRESSION_STAGES.COMPARISON
                    tokenizer.skip_white_space()
            elif current_stage == EXPRESSION_STAGES.KEY_VALUE:
                # job_id =          :id
                # job_id =          :id and  ...
                # pk=p and          x=y
                # pk=p and fn(x, y1, y1 )
                #                      ^ --> ^
                key_values.append(
                    expression_attribute_values.get(
                        current_phrase, {"S": current_phrase}
                    )
                )
                current_phrase = ""
                if comparison.upper() != "BETWEEN" or len(key_values) == 2:
                    results.append((key_name, comparison, key_values))
                    key_values = []
                tokenizer.skip_white_space()
                if tokenizer.peek() == ")":
                    tokenizer.skip_characters(")")
                    current_stage = EXPRESSION_STAGES.EOF
                    break
                elif tokenizer.is_eof():
                    break
                tokenizer.skip_characters("AND", case_sensitive=False)
                tokenizer.skip_white_space()
                if comparison.upper() == "BETWEEN":
                    # We can expect another key_value, i.e. BETWEEN x and y
                    # We should add some validation, to not allow BETWEEN x and y and z and ..
                    pass
                else:
                    current_stage = EXPRESSION_STAGES.INITIAL_STAGE
            elif current_stage == EXPRESSION_STAGES.COMPARISON:
                # hashkey = :id and sortkey       =      :sk
                # hashkey = :id and sortkey BETWEEN      x and y
                #                                  ^ --> ^
                comparison = current_phrase
                current_phrase = ""
                current_stage = EXPRESSION_STAGES.KEY_VALUE
            continue
        if crnt_char in ["=", "<", ">"] and current_stage in [
            EXPRESSION_STAGES.KEY_NAME,
            EXPRESSION_STAGES.INITIAL_STAGE,
            EXPRESSION_STAGES.COMPARISON,
        ]:
            if current_stage in [
                EXPRESSION_STAGES.KEY_NAME,
                EXPRESSION_STAGES.INITIAL_STAGE,
            ]:
                if expression_attribute_names.get(current_phrase):
                    key_name = expression_attribute_names[current_phrase]
                    expression_attribute_names_used.append(current_phrase)
                else:
                    key_name = current_phrase
            current_phrase = ""
            if crnt_char in ["<", ">"] and tokenizer.peek() == "=":
                comparison = crnt_char + tokenizer.__next__()
            else:
                comparison = crnt_char
            tokenizer.skip_white_space()
            current_stage = EXPRESSION_STAGES.KEY_VALUE
            continue
        if crnt_char in [","]:
            if current_stage == EXPRESSION_STAGES.KEY_NAME:
                # hashkey = :id and begins_with(sortkey,     :sk)
                #                                      ^ --> ^
                if expression_attribute_names.get(current_phrase):
                    key_name = expression_attribute_names[current_phrase]
                    expression_attribute_names_used.append(current_phrase)
                else:
                    key_name = current_phrase
                current_phrase = ""
                current_stage = EXPRESSION_STAGES.KEY_VALUE
                tokenizer.skip_white_space()
                continue
            else:
                raise MockValidationException(
                    f'Invalid KeyConditionExpression: Syntax error; token: "{current_phrase}"'
                )
        if crnt_char in [")"]:
            if current_stage == EXPRESSION_STAGES.KEY_VALUE:
                # hashkey = :id and begins_with(sortkey, :sk)
                #                                            ^
                value = expression_attribute_values.get(current_phrase, current_phrase)
                current_phrase = ""
                key_values.append(value)
                results.append((key_name, comparison, key_values))
                key_values = []
                tokenizer.skip_white_space()
                if tokenizer.is_eof() or tokenizer.peek() == ")":
                    break
                else:
                    tokenizer.skip_characters("AND", case_sensitive=False)
                    tokenizer.skip_white_space()
                    current_stage = EXPRESSION_STAGES.INITIAL_STAGE
                    continue
        if crnt_char in [""]:
            # hashkey =                   :id
            # hashkey = :id and sortkey = :sk
            #                                ^
            if current_stage == EXPRESSION_STAGES.KEY_VALUE:
                if current_phrase not in expression_attribute_values:
                    raise MockValidationException(
                        "Invalid condition in KeyConditionExpression: Multiple attribute names used in one condition"
                    )
                key_values.append(expression_attribute_values[current_phrase])
                results.append((key_name, comparison, key_values))
                break
        if crnt_char == "(":
            # hashkey = :id and begins_with(      sortkey,     :sk)
            #                              ^ --> ^
            # (hash_key = :id) and (sortkey = :sk)
            #                     ^
            if current_stage in [EXPRESSION_STAGES.INITIAL_STAGE]:
                if not current_phrase:
                    # hashkey = :id and (begins_with(sortkey, :sk))
                    #                   ^
                    continue
                if current_phrase not in ["begins_with", ""]:
                    raise MockValidationException(
                        f"Invalid KeyConditionExpression: Invalid function name; function: {current_phrase}"
                    )
                comparison = current_phrase
                current_phrase = ""
                tokenizer.skip_white_space()
                current_stage = EXPRESSION_STAGES.KEY_NAME
                continue
            if current_stage is None:
                # (hash_key = :id .. )
                # ^
                continue

        current_phrase += crnt_char
        if current_stage is None:
            current_stage = EXPRESSION_STAGES.INITIAL_STAGE

    hash_key_conditions, range_key_conditions = validate_schema(results, schema)

    return (
        hash_key_conditions,
        range_key_conditions,
        expression_attribute_names_used,
    )


# Validate that the schema-keys are encountered in our query
def validate_schema(
    results: Any, schema: list[dict[str, str]]
) -> tuple[
    list[tuple[str, dict[str, Any]]],
    list[tuple[str, str, list[dict[str, Any]]]],
]:
    """
    Validate query conditions against the schema and extract key values.

    For multi-attribute keys:
    - ALL hash keys must be present with equality (=)
    - Range keys must be specified left-to-right (cannot skip)
    - Only the LAST range key in the query can use non-equality operators
    - Earlier range keys in the query must use equality (=)

    Returns:
        - hash_key_conditions: List of (attr_name, value) for all hash keys
        - range_key_conditions: List of (attr_name, comparison, values) for range keys
    """
    # Build a lookup from results: {key_name: (comparison, values)}
    results_by_key: dict[str, tuple[str, list[dict[str, Any]]]] = {
        key: (comparison, values) for key, comparison, values in results
    }

    # Get all hash and range keys from schema in order
    hash_keys = get_keys(schema, "HASH")
    range_keys = get_keys(schema, "RANGE")
    provided_keys = set(results_by_key.keys())
    schema_keys = set(hash_keys + range_keys)

    # === Validate HASH keys ===
    # All hash keys must be present with equality
    # Check missing hash keys FIRST (before checking for invalid keys)
    hash_key_conditions: list[tuple[str, dict[str, Any]]] = []

    for hash_key in hash_keys:
        if hash_key not in results_by_key:
            raise MockValidationException(
                f"Query condition missed key schema element: {hash_key}"
            )
        comparison, values = results_by_key[hash_key]
        if comparison != "=":
            raise MockValidationException("Query key condition not supported")
        value = values[0]
        if "S" in value and value["S"] == "":
            raise KeyIsEmptyStringException(hash_key)  # type: ignore[arg-type]
        hash_key_conditions.append((hash_key, value))

    # === Validate RANGE keys ===
    # Range keys must be specified left-to-right, only last can use non-equality
    range_key_conditions: list[tuple[str, str, list[dict[str, Any]]]] = []

    # Find which range keys are in the query (in schema order)
    range_keys_in_query: list[str] = [k for k in range_keys if k in results_by_key]

    # Check for unknown keys (keys in query but not in schema)
    unknown_keys = provided_keys - schema_keys
    if unknown_keys:
        # User provided keys not in schema - report first missing range key if any,
        # otherwise report generic error
        missing_range_keys = [k for k in range_keys if k not in results_by_key]
        if missing_range_keys:
            raise MockValidationException(
                f"Query condition missed key schema element: {missing_range_keys[0]}"
            )
        raise MockValidationException("Query key condition not supported")

    for i, range_key in enumerate(range_keys):
        if range_key not in results_by_key:
            # This range key is not in the query
            # Check if any later range keys ARE in the query (would be skipping)
            later_keys_in_query = [
                k for k in range_keys[i + 1 :] if k in results_by_key
            ]
            if later_keys_in_query:
                # User skipped this key but specified a later one - error
                raise MockValidationException(
                    f"RANGE key attributes {range_key} must have equality conditions "
                    f"specified in the query because a condition is present on key "
                    f"attribute {later_keys_in_query[0]}"
                )
            # No more range keys in query, we're done
            break

        comparison, values = results_by_key[range_key]
        is_last_in_query = range_key == range_keys_in_query[-1]

        if is_last_in_query:
            # Last range key in query can use any comparison
            if {"S": ""} in values:
                raise KeyIsEmptyStringException(range_key)
            range_key_conditions.append((range_key, comparison.upper(), values))
        else:
            # Not last in query - must be equality
            if comparison != "=":
                raise MockValidationException("Query key condition not supported")
            value = values[0]
            if "S" in value and value["S"] == "":
                raise KeyIsEmptyStringException(range_key)  # type: ignore[arg-type]
            range_key_conditions.append((range_key, "=", [value]))

    return hash_key_conditions, range_key_conditions



from moto.core.exceptions import ServiceException


class AutoscalingClientError(ServiceException):
    pass


class ResourceContentionError(AutoscalingClientError):
    code = "ResourceContention"
    message = "You already have a pending update to an Auto Scaling resource (for example, a group, instance, or load balancer)."


class ValidationError(AutoscalingClientError):
    code = "ValidationError"


class InvalidInstanceError(ValidationError):
    def __init__(self, instance_id: str):
        super().__init__(f"Instance [{instance_id}] is invalid.")



"""AWS Config integration for SNS resources."""

import json
from typing import Any, Optional

from moto.core.common_models import ConfigQueryModel
from moto.sns import sns_backends
from moto.sns.models import SNSBackend


class SNSConfigQuery(ConfigQueryModel[SNSBackend]):
    """Config query model for SNS topics."""

    def list_config_service_resources(
        self,
        account_id: str,
        partition: str,
        resource_ids: Optional[list[str]],
        resource_name: Optional[str],
        limit: int,
        next_token: Optional[str],
        backend_region: Optional[str] = None,
        resource_region: Optional[str] = None,
        aggregator: Optional[dict[str, Any]] = None,
    ) -> tuple[list[dict[str, Any]], Optional[str]]:
        region = resource_region or backend_region or "us-east-1"
        backend = self.backends[account_id][region]

        topics, new_token = backend.list_config_service_resources(
            resource_ids=resource_ids,
            resource_name=resource_name,
            limit=limit,
            next_token=next_token,
        )

        return topics, new_token

    def get_config_resource(
        self,
        account_id: str,
        partition: str,
        resource_id: str,
        resource_name: Optional[str] = None,
        backend_region: Optional[str] = None,
        resource_region: Optional[str] = None,
    ) -> Optional[dict[str, Any]]:
        region = resource_region or backend_region or "us-east-1"
        backend = self.backends[account_id][region]

        config_data = backend.get_config_resource(resource_id)

        if not config_data:
            return None

        if resource_name and config_data.get("resourceName") != resource_name:
            return None

        if "configuration" in config_data and not isinstance(
            config_data["configuration"], str
        ):
            config_data["configuration"] = json.dumps(config_data["configuration"])

        if "supplementaryConfiguration" in config_data:
            for field, value in config_data["supplementaryConfiguration"].items():
                if not isinstance(value, str):
                    config_data["supplementaryConfiguration"][field] = json.dumps(value)

        return config_data


sns_config_query = SNSConfigQuery(sns_backends)



from typing import Any

from moto.core.exceptions import RESTError, ServiceException

ERROR_WITH_ACCESS_POINT_NAME = """{% extends 'wrapped_single_error' %}
{% block extra %}<AccessPointName>{{ name }}</AccessPointName>{% endblock %}
"""


ERROR_WITH_ACCESS_POINT_POLICY = """{% extends 'wrapped_single_error' %}
{% block extra %}<AccessPointName>{{ name }}</AccessPointName>{% endblock %}
"""

ERROR_WITH_MRAP_NAME = """{% extends 'wrapped_single_error' %}
{% block extra %}<Name>{{ name }}</Name>{% endblock %}
"""


ERROR_WITH_REQUEST_TOKEN = """{% extends 'wrapped_single_error' %}
{% block extra %}<RequestTokenARN>{{ request_token }}</RequestTokenARN>{% endblock %}
"""


class S3ControlError(RESTError):
    extended_templates = {
        "ap_not_found": ERROR_WITH_ACCESS_POINT_NAME,
        "apf_not_found": ERROR_WITH_ACCESS_POINT_POLICY,
        "mrap_not_found": ERROR_WITH_MRAP_NAME,
        "mrap_policy_not_found": ERROR_WITH_MRAP_NAME,
        "mrap_operation_not_found": ERROR_WITH_REQUEST_TOKEN,
    }
    env = RESTError.extended_environment(extended_templates)

    def __init__(self, *args: Any, **kwargs: Any):
        kwargs.setdefault("template", "single_error")
        super().__init__(*args, **kwargs)


class AccessPointNotFound(S3ControlError):
    code = 404

    def __init__(self, name: str, **kwargs: Any):
        kwargs.setdefault("template", "ap_not_found")
        kwargs["name"] = name
        super().__init__(
            "NoSuchAccessPoint", "The specified accesspoint does not exist", **kwargs
        )


class AccessPointPolicyNotFound(S3ControlError):
    code = 404

    def __init__(self, name: str, **kwargs: Any):
        kwargs.setdefault("template", "apf_not_found")
        kwargs["name"] = name
        super().__init__(
            "NoSuchAccessPointPolicy",
            "The specified accesspoint policy does not exist",
            **kwargs,
        )


class MultiRegionAccessPointNotFound(S3ControlError):
    code = 404

    def __init__(self, name: str, **kwargs: Any):
        kwargs.setdefault("template", "mrap_not_found")
        kwargs["name"] = name
        super().__init__(
            "NoSuchMultiRegionAccessPoint",
            "The specified multi-region access point does not exist",
            **kwargs,
        )


class MultiRegionAccessPointPolicyNotFound(S3ControlError):
    code = 404

    def __init__(self, name: str, **kwargs: Any):
        kwargs.setdefault("template", "mrap_policy_not_found")
        kwargs["name"] = name
        super().__init__(
            "NoSuchMultiRegionAccessPointPolicy",
            "The specified multi-region access point policy does not exist",
            **kwargs,
        )


class MultiRegionAccessPointOperationNotFound(S3ControlError):
    code = 404

    def __init__(self, request_token: str, **kwargs: Any):
        kwargs.setdefault("template", "mrap_operation_not_found")
        kwargs["request_token"] = request_token
        super().__init__(
            "NoSuchAsyncRequest",
            "The specified async request does not exist",
            **kwargs,
        )


class NoSuchPublicAccessBlockConfiguration(ServiceException):
    # Note that this exception is in the different format then the S3 exception with the same name
    # This exception should return a nested response `<ErrorResponse><Error>..`
    # The S3 variant uses a flat `<Error>`-response
    code = "NoSuchPublicAccessBlockConfiguration"

    def __init__(self) -> None:
        super().__init__("The public access block configuration was not found")


class InvalidRequestException(S3ControlError):
    code = 400

    def __init__(self, message: str, **kwargs: Any):
        super().__init__(
            "InvalidRequest",
            message,
            **kwargs,
        )


class StorageLensConfigurationNotFound(S3ControlError):
    code = 404

    def __init__(self, config_id: str, **kwargs: Any):
        super().__init__(
            "NoSuchConfiguration",
            f"The specified configuration does not exist: {config_id}",
            **kwargs,
        )



from datetime import datetime, timezone
from typing import Optional, TypedDict


class ManagedPermissionDict(TypedDict):
    arn: str
    version: str
    defaultVersion: bool
    name: str
    resourceType: str
    status: str
    creationTime: str
    lastUpdatedTime: str
    isResourceTypeDefault: bool
    permissionType: str


def format_ram_permission(
    name: str,
    resource_type: str,
    version: str = "1",
    arn_prefix: str = "arn:aws:ram::aws:permission/",
    status: str = "ATTACHABLE",
    creation_time: Optional[str] = None,
    last_updated_time: Optional[str] = None,
    is_resource_type_default: bool = True,
    permission_type: str = "AWS_MANAGED",
    default_version: bool = True,
) -> ManagedPermissionDict:
    """
    Format a RAM (Resource Access Manager) permission dictionary with the
    specified attributes.

    Args:
        name (str): The name of the permission.
        resource_type (str): The type of resource the permission applies to.
        version (str): The version of the permission. Defaults to "1".
        arn_prefix (str): The prefix for the permission ARN. Defaults
            to "arn:aws:ram::aws:permission/".
        status (str): The status of the permission. Defaults to
            "ATTACHABLE".
        creation_time (str, optional): The creation time in UTC. If not
            provided, uses the current time.
        last_updated_time (str, optional): The last updated time in UTC. If
            not provided, uses the creation time.
        is_resource_type_default (bool): Whether this is the default
            permission for the resource type. Defaults to True.
        permission_type (str): The type of permission (e.g.,
            "AWS_MANAGED"). Defaults to "AWS_MANAGED".
        default_version (bool): Whether this is the default version.
            Defaults to True.

    Returns:
        ManagedPermissionDict: A dictionary representing the formatted RAM permission.
    """
    now = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
    creation_time = creation_time or now
    last_updated_time = last_updated_time or creation_time
    arn = f"{arn_prefix}{name}"
    return {
        "arn": arn,
        "version": version,
        "defaultVersion": default_version,
        "name": name,
        "resourceType": resource_type,
        "status": status,
        "creationTime": creation_time,
        "lastUpdatedTime": last_updated_time,
        "isResourceTypeDefault": is_resource_type_default,
        "permissionType": permission_type,
    }


# List of tuples with resourceType, serviceName, and resourceRegionScope based
# on the SHARED_RESOURCE_TYPES in models.py
_RESOURCE_TYPE_DEFINITIONS = [
    ("rds:Cluster", "rds", "REGIONAL"),
    ("imagebuilder:Component", "imagebuilder", "REGIONAL"),
    ("networkmanager:CoreNetwork", "networkmanager", "GLOBAL"),
    ("resource-groups:Group", "resource-groups", "REGIONAL"),
    ("imagebuilder:Image", "imagebuilder", "REGIONAL"),
    ("imagebuilder:ImageRecipe", "imagebuilder", "REGIONAL"),
    ("license-manager:LicenseConfiguration", "license-manager", "REGIONAL"),
    ("appmesh:Mesh", "appmesh", "REGIONAL"),
    ("ec2:PrefixList", "ec2", "REGIONAL"),
    ("codebuild:Project", "codebuild", "REGIONAL"),
    ("codebuild:ReportGroup", "codebuild", "REGIONAL"),
    ("route53resolver:ResolverRule", "route53resolver", "REGIONAL"),
    ("ec2:Subnet", "ec2", "REGIONAL"),
    ("ec2:TransitGatewayMulticastDomain", "ec2", "REGIONAL"),
    ("glue:Database", "glue", "REGIONAL"),
    ("glue:Table", "glue", "REGIONAL"),
    ("glue:Catalog", "glue", "REGIONAL"),
]

# List of dictionaries representing RAM resource types based on the definitions
RAM_RESOURCE_TYPES = [
    {
        "resourceType": resource_type,
        "serviceName": service_name,
        "resourceRegionScope": region_scope,
    }
    for resource_type, service_name, region_scope in _RESOURCE_TYPE_DEFINITIONS
]

# List of AWS managed permissions for RAM, formatted using the
# format_ram_permission function.
AWS_MANAGED_PERMISSIONS = [
    format_ram_permission(
        name="AWSRAMDefaultPermissionRDSCluster",
        resource_type="rds:Cluster",
        creation_time="2022-06-30 17:04:16.671000",
    ),
    format_ram_permission(
        name="AWSRAMDefaultPermissionImageBuilderComponent",
        resource_type="imagebuilder:Component",
        creation_time="2022-06-30 17:04:21.977000",
    ),
    format_ram_permission(
        name="AWSRAMDefaultPermissionsNetworkManagerCoreNetwork",
        resource_type="networkmanager:CoreNetwork",
        version="3",
        creation_time="2024-10-23 16:34:51.604000",
    ),
    format_ram_permission(
        name="AWSRAMTransitGatewayPermissionsNetworkManagerCoreNetwork",
        resource_type="networkmanager:CoreNetwork",
        is_resource_type_default=False,
        creation_time="2022-06-30 17:03:48.071000",
    ),
    format_ram_permission(
        name="AWSRAMVPCPermissionsNetworkManagerCoreNetwork",
        resource_type="networkmanager:CoreNetwork",
        is_resource_type_default=False,
        creation_time="2022-06-30 17:03:46.477000",
    ),
    format_ram_permission(
        name="AWSRAMDefaultPermissionResourceGroup",
        resource_type="resource-groups:Group",
        creation_time="2022-06-30 17:04:10.914000",
    ),
    format_ram_permission(
        name="AWSRAMDefaultPermissionImageBuilderImage",
        resource_type="imagebuilder:Image",
        creation_time="2022-06-30 17:04:21.201000",
    ),
    format_ram_permission(
        name="AWSRAMDefaultPermissionImageBuilderImageRecipe",
        resource_type="imagebuilder:ImageRecipe",
        creation_time="2022-06-30 17:04:22.832000",
    ),
    format_ram_permission(
        name="AWSRAMDefaultPermissionLicenseConfiguration",
        resource_type="license-manager:LicenseConfiguration",
        creation_time="2024-01-31 16:36:13.664000",
    ),
    format_ram_permission(
        name="AWSRAMDefaultPermissionAppMesh",
        resource_type="appmesh:Mesh",
        version="3",
        creation_time="2023-03-28 12:27:55.910000",
    ),
    format_ram_permission(
        name="AWSRAMDefaultPermissionPrefixList",
        resource_type="ec2:PrefixList",
        creation_time="2022-06-30 17:04:20.594000",
    ),
    format_ram_permission(
        name="AWSRAMDefaultPermissionCodeBuildProject",
        resource_type="codebuild:Project",
        creation_time="2022-06-30 17:04:17.595000",
    ),
    format_ram_permission(
        name="AWSRAMDefaultPermissionCodeBuildReportGroup",
        resource_type="codebuild:ReportGroup",
        creation_time="2022-06-30 17:04:17.761000",
    ),
    format_ram_permission(
        name="AWSRAMDefaultPermissionResolverRule",
        resource_type="route53resolver:ResolverRule",
        creation_time="2022-06-30 17:04:14.752000",
    ),
    format_ram_permission(
        name="AWSRAMDefaultPermissionSubnet",
        resource_type="ec2:Subnet",
        creation_time="2024-10-30 15:11:41.358000",
    ),
    format_ram_permission(
        name="AWSRAMDefaultPermissionTransitGatewayMulticastDomain",
        resource_type="ec2:TransitGatewayMulticastDomain",
        creation_time="2022-06-30 17:04:13.461000",
    ),
    format_ram_permission(
        name="AWSRAMDefaultPermissionGlueDatabase",
        resource_type="glue:Database",
        creation_time="2022-06-30 17:04:25.912000",
    ),
    format_ram_permission(
        name="AWSRAMLFEnabledGlueAllTablesReadWriteForDatabase",
        resource_type="glue:Database",
        is_resource_type_default=False,
        creation_time="2023-06-27 17:05:24.471000",
    ),
    format_ram_permission(
        name="AWSRAMLFEnabledGlueDatabaseReadWrite",
        resource_type="glue:Database",
        is_resource_type_default=False,
        creation_time="2023-06-27 17:05:27.153000",
    ),
    format_ram_permission(
        name="AWSRAMPermissionGlueAllTablesReadWriteForDatabase",
        resource_type="glue:Database",
        version="3",
        is_resource_type_default=False,
        creation_time="2022-10-27 14:19:43.345000",
    ),
    format_ram_permission(
        name="AWSRAMPermissionGlueDatabaseReadWrite",
        resource_type="glue:Database",
        version="3",
        is_resource_type_default=False,
        creation_time="2022-10-27 14:20:08.415000",
    ),
    format_ram_permission(
        name="AWSRAMPermissionGlueTableReadWriteForDatabase",
        resource_type="glue:Database",
        version="3",
        is_resource_type_default=False,
        creation_time="2022-10-27 14:19:17.062000",
    ),
    format_ram_permission(
        name="AWSRAMPermissionLFTagGlueDatabaseReadWrite",
        resource_type="glue:Database",
        is_resource_type_default=False,
        creation_time="2022-10-27 14:22:14.138000",
    ),
    format_ram_permission(
        name="AWSRAMPermissionLFTagGlueTableReadWriteForDatabase",
        resource_type="glue:Database",
        is_resource_type_default=False,
        creation_time="2022-10-27 14:21:49.920000",
    ),
    format_ram_permission(
        name="AWSRAMDefaultPermissionGlueTable",
        resource_type="glue:Table",
        creation_time="2022-06-30 17:04:24.372000",
    ),
    format_ram_permission(
        name="AWSRAMLFEnabledGlueDatabaseReadWriteForTable",
        resource_type="glue:Table",
        is_resource_type_default=False,
        creation_time="2023-06-27 17:05:33.534000",
    ),
    format_ram_permission(
        name="AWSRAMLFEnabledGlueTableReadWrite",
        resource_type="glue:Table",
        is_resource_type_default=False,
        creation_time="2023-06-27 17:05:21.095000",
    ),
    format_ram_permission(
        name="AWSRAMPermissionGlueDatabaseReadWriteForTable",
        resource_type="glue:Table",
        version="2",
        is_resource_type_default=False,
        creation_time="2022-06-30 17:04:09.034000",
    ),
    format_ram_permission(
        name="AWSRAMPermissionGlueTableReadWrite",
        resource_type="glue:Table",
        version="2",
        is_resource_type_default=False,
        creation_time="2022-06-30 17:04:03.947000",
    ),
    format_ram_permission(
        name="AWSRAMPermissionLFTagGlueDatabaseReadWriteForTable",
        resource_type="glue:Table",
        is_resource_type_default=False,
        creation_time="2022-10-27 14:22:26.191000",
    ),
    format_ram_permission(
        name="AWSRAMPermissionLFTagGlueTableReadWrite",
        resource_type="glue:Table",
        is_resource_type_default=False,
        creation_time="2022-10-27 14:22:01.908000",
    ),
    format_ram_permission(
        name="AWSRAMDefaultPermissionGlueCatalog",
        resource_type="glue:Catalog",
        creation_time="2022-06-30 17:04:26.612000",
    ),
    format_ram_permission(
        name="AWSRAMPermissionGlueAllTablesReadWriteForCatalog",
        resource_type="glue:Catalog",
        version="3",
        is_resource_type_default=False,
        creation_time="2022-10-27 14:19:56.250000",
    ),
    format_ram_permission(
        name="AWSRAMPermissionGlueDatabaseReadWriteForCatalog",
        resource_type="glue:Catalog",
        version="3",
        is_resource_type_default=False,
        creation_time="2022-10-27 14:20:20.322000",
    ),
    format_ram_permission(
        name="AWSRAMPermissionGlueTableReadWriteForCatalog",
        resource_type="glue:Catalog",
        version="3",
        is_resource_type_default=False,
        creation_time="2022-10-27 14:19:30.644000",
    ),
    format_ram_permission(
        name="AWSRAMPermissionLFTagGlueDatabaseReadWriteForCatalog",
        resource_type="glue:Catalog",
        is_resource_type_default=False,
        creation_time="2022-10-27 14:22:37.435000",
    ),
    format_ram_permission(
        name="AWSRAMPermissionLFTagGlueTableReadWriteForCatalog",
        resource_type="glue:Catalog",
        is_resource_type_default=False,
        creation_time="2022-10-27 14:21:37.593000",
    ),
]
